//                              -*- Mode: C++ -*-
//:ECALDetector.hh
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 1/10/97
// Last Update ........
//
//
#ifndef ECALDETECTOR_HH_
#define ECALDETECTOR_HH_

#include "care/care.hh"
#include "care/caredata.hh"
#include <vector>

#ifdef CMATCH
#  ifdef RTRALINK
     class RcTraPoint;
     typedef RcTraPoint TrackInfo; 
#  else
     class RcSegPoint;
     typedef RcSegPoint TrackInfo; 
#  endif
#else
  class   TrackInfo;
#endif

class ECBuilder;
class EMCluster;
class EMClusterTree;
class ECHitCellArray;
/**
class ECALDetector.
  Provide a collection of basic function for the ECAL detector 
  reconstruction and analysis.
  
Patterns: Singleton, Facade
  */

class ECALDetector {
protected:
  //--
  /// Constructor
  ECALDetector();
  /// disallow copy constructor
  ECALDetector(const ECALDetector &){};
  
public:
 
  /**@name Beheavioral methods */
  //@{ 
  //--
  /** Prepare internal data for analyzing the events. 
   */
  static int initialize();
  //--
  /// Find and return shower hypotheses
  const std::vector<const EMCluster*> & findShowers(EcalEvent_t * start);
  //--
  /// Find and return shower hypotheses
  const std::vector<const EMCluster*> & findShowers();
  //--
  /// Find and return all signals detected on ECAL
  const std::vector<const EMCluster*> & findAllSignals();
  //--
  /// Store the neutral clusters found in the RTRA bank
  int storeNeutrals();
  //--
  /// Stop of ECAL reconstruction.
  static int stop();  
  //--
  /// First step of ECAL reconstruction. Same as findShowers().
  static int careevnt1();  
  //--
  /// Second step of ECAL reconstruction. Same as findAllSignals().
  static int careevnt2();  
  //--
  /// Third step of ECAL reconstruction. 
  static int careevnt3(); // No action for the moment !!!  
  //--
  /// Store parameters of a track
  static void storeTrack(TrackInfo & trackInfo);
  //--
  /// get the information for the new DB data
  static int newDBData(int run = 0); 
  //--
  /// get the information for the new DB data
  static int newDBData(const char * dbname, int * major, int * minor);
  //@}

  /**@name access methods */
  //@{ 
  //--
  /// Return the instance
  inline static ECALDetector* instance();
  //--
  /// Return shower hypotheses
  inline const std::vector<const EMCluster*> & showers() const;
  //--
  /// Return signals clusters
  inline const std::vector<const EMCluster*> & signals() const;
  //--
  /// Return top cluster
  inline const EMClusterTree & TopCluster() const ;
  //--
  /// Return geometry informations
  static ECAL_Info & Infos(){ return ECalInfo; }
  ///--
  // Return the dbflag;
  inline static int getDBflag();
  ///--
  // Set the db flag
  inline static void setDBflag( int fl);
  ///--
  // Get the event header for the first event
  inline const EVHD & evheader() const; 
  //@}  
private:
  /// The only instance
  static ECALDetector * ecalDetector;
  /// the Cluster Tree.
  EMClusterTree *       topClusterTree;
  /// the Builder of the tree.
  ECBuilder *           clusBuilder;
  /// array of Hit Cells
  ECHitCellArray *      pSignalCells;
  /// to store the shower hypotheses
  std::vector<const EMCluster*> showerArray;         
  /// to store the signal (=showers+tracks) hypotheses
  std::vector<const EMCluster*> signalArray;         

  EVHD  evhd; // event header of the first event;
};

// Return the instance
ECALDetector* ECALDetector::instance(){
  if(ecalDetector==0){
    initialize();
  }
  return ecalDetector;
}

//--
// Return shower hypotheses
const std::vector<const EMCluster*> & ECALDetector::showers() const {
  return showerArray;
}

//--
// Return signals clusters
const std::vector<const EMCluster*> & ECALDetector::signals() const { 
  return signalArray;
}

//--
// Return top cluster
const EMClusterTree & ECALDetector::TopCluster() const {
  return *topClusterTree;
}
//--

// get the event header
const EVHD & ECALDetector::evheader() const { return evhd;}

#endif // ECALDETECTOR_HH_
