//                              -*- Mode: C++ -*-
//: ESEnCorStrategy.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 25/10/99
// Last Update ........
//

#ifndef ESENCORSTRATEGY_HH_
#define ESENCORSTRATEGY_HH_

#include "care/EAClusterAnalyzer.hh"
#include "care/ESPosition.hh"
/**
class ESEnCorStrategy.
  The most general algorithm to correct the electromagnetic
  energy measured in a shower.
  Pure abstract class.

Usage:  
       <PRE>
       // Suppose you have an EMCluster* initialized somehow
       const EMCluster * cluster = ...

       // Prepare somehow the corrected COG values
       ESPosition corcog = ....
       // Create a position strategy with the StrategyFactory class
       StrategyFactory SF;

       // Ask for an energy correction strategy
       ESEnCorStrategy* enCorStra = SF.EnergyCorStrategy()
       // now evaluate the correction of the cluster
       enCorStra->evaluateEnergy( cluster , corcog );
       // you can now access the data with
       ESPosition data = posStra->getData();
       </PRE>
      
*/

class ESEnCorStrategy : public EAClusterAnalyzer {
public:
 
  /// Constructors
  ESEnCorStrategy();
  //-- 
  /// Destructor: currently void but needs to be virtual
  virtual ~ESEnCorStrategy();
  
  //-- Behavioural operations
  ///   evaluate the energy correction for a cluster. To be defined by subclasses.
  virtual Energy_t evaluateEnergy(const EMCluster* clus, ESPosition & positionData);
  
  //   Access functions 
  //--
  ///   Inquire for the energy stored
  inline Energy_t  getEnergy() const;
  ///   Inquire for the error on energy (squared)
  inline Energy_t  getSqrError() const;
  ///   get the position data
  inline ESPosition & getPosData() const;
  ///   get correction factor
  inline float     getFactor() const;

protected:
  //              Data Part 
  //--
  /// The corrected energy and its error
  Energy_t energy;
  Energy_t energyVariance;
  /// The position data
  ESPosition * posData; 
  /// the correction factor
  float corrfact;
};


// inlines Access functions 

///   Inquire for the energy stored
Energy_t  ESEnCorStrategy::getEnergy() const{
  return energy;
}

///   Inquire for the error on energy (squared)
Energy_t  ESEnCorStrategy::getSqrError() const {
  return energyVariance;
}

///   get the position data
ESPosition & ESEnCorStrategy::getPosData() const {
  return *posData;
}

///   Inquire for the correction factor
float ESEnCorStrategy::getFactor() const {
  return corrfact;
}


#endif  //   ESENPOSSTRATEGY_HH_



