/* Defines 
   1) iteration through all cells, sections, chambers or setups in the 
   detector. Note that the types of different Setups can be different.
   Usage: 
     for_each_cell   (action, detector);
     for_each_section(action, detector);
     for_each_chamber(action, detector);

     for_each_setup  (action, detector);

   Here in case of for_each_cell /section,chamber/ (...) methods
     action(icell); /action(isection); action(ichamber);/
   function object should be defined for all Setup types in the detector.
   In case of for_each_setup
     action(setup);
   should be defined.

   (See also setup_tools.hh for an explanation of for_each_...(..)
    mechanism).

   There is one more function:
     for_each_setup_with_name(action, detector);
   available. It differs from for_each_setup(action, detector); in that
   it supplies for the action not only setup, but also its name in
   the detector. That is the 
     action(setup, setup_name);
   function object should be defined in this case instead of
   action(setup). 
   The setup name can be needed for some algorithms (see also the
   discussion in setup_tools.hh why, e.g. in for_each_cell(action) action
   should work with ICell, not the Cell itself). 
   Here both versions are provided.

......................................................................
   Example how to print the number of chambers in setups:
   struct PrintNumberOfChambers {
     template <class AnySetup>
     void operator()(AnySetup& setup, const string& setup_name) {
       cout << "Setup " << setup_name ": " << setup.size() << '\n';
     }
   };
   for_each_setup_with_name(PrintNumberOfChambers(), detector);
......................................................................

   2) castToSetupPtr(const Detector& detector, const string& setup_name,
			   AnySetup** ptr);
   or:
      castToSetupPtr(__SetupBase* psetup_base, AnySetup** ptr);}
   
   Can be used for dynamic casting to AnySetup type. It assumes that
   Setup with "setup_name" already exists in the detector (for the 1st 
   version) and has the right type AnySetup (for both versions).
   Otherwise global_error() is set.
   */

// Author V.Balagura (balagura@mail.desy.de)   (18.02.99).

#ifndef _detector_tools_hh
#define _detector_tools_hh 1

#include "ptset/detector.hh"
#include "ptset/setup_tools.hh"

/* -------------------- for_each_... -------------------- */
template<class Action>
inline void for_each_setup(const Action& action, Detector& det) {
  for ( Detector::iterator id = det.begin(); id != det.end(); ++id )
    apply_virtually(action)( ((*id).second)() );
}
template <class Action>
inline void for_each_cell(const Action& action, Detector& det) {
  for_each_setup(for_each_cell(action), det);
}
template <class Action>
inline void for_each_section(const Action& action, Detector& det) {
  for_each_setup(for_each_section(action), det);
}
template <class Action>
inline void for_each_chamber(const Action& action, Detector& det) {
  for_each_setup(for_each_chamber(action), det);
}

template <class Action>
struct __ActionWithName {
  Action action;
  const string& setup_name; 
  __ActionWithName(const Action& act, const string& name)
    : action(act), setup_name(name) {}
  template <class AnySetup>
  void operator()(AnySetup& setup) { action(setup, setup_name); }
};
template<class Action>
inline void for_each_setup_with_name(const Action& action, 
				       Detector& det) {
  for ( Detector::iterator id = det.begin(); id != det.end(); ++id )
    apply_virtually(__ActionWithName<Action>(action, (*id).first))
      ( ((*id).second)() );
}
/* -------------------- castToSetupPtr -------------------- */
template <class AnySetup>
class __CastToSetupPtr {
  AnySetup** ptr;
public:
  __CastToSetupPtr(AnySetup** ptr_) : ptr(ptr_) {}

  void operator()(AnySetup& setup) const { *ptr = &setup; }
  template <class AnyOtherSetup>
  void operator()(AnyOtherSetup&)  const { 
    //    *ptr = 0; 
    global_error()("wrong type conversion in castToSetupPtr()");
  }
};

template <class AnySetup>
inline void castToSetupPtr(__SetupBase* psetup_base, AnySetup** ptr) {
  apply_virtually(__CastToSetupPtr<AnySetup>(ptr))(psetup_base);
}
template <class AnySetup>
inline void castToSetupPtr(Detector& det, const string& setup_name,
			   AnySetup** ptr) {
  apply_virtually(__CastToSetupPtr<AnySetup>(ptr))(det, setup_name);
}

#endif
