/* The function
      make_chamber_from_rectangles
      (chamber,
       vector<ChamberRectangularPartDescription>* rectangles);
   makes the following. Based on the chamber structure it divides it into
   non overlapping rectangles. Inside each rectangle the structure of cells
   is uniform. There are no holes in such rectangle and the grid of cells
   is the same within the rectangle.
   This function can be used, e.g. to obtain the rectangles needed to define
   the geometry in ARTE (HERA-B) geometry banks.
   The rectangle is represented by the following information:
     struct ChamberRectangularPartDescription {
       Index ind1, ind2;
       Grid grid;
     };
   The Grid grid data field contains the cells grid. The rectangle boundaries
   are passing through the nodes of the grid (not the centers of the cells).
   Indexes ind1 and ind2 show the low-left and upper-right corners of the
   rectangle respectively.
   Thus all cells inside the rectangle have
   ind1.ix <= ind.ix < ind2.ix and
   ind1.iy <= ind.iy < ind2.iy.
   (Upper-right corner is never reached).
   To obtain the rectangle itself one can write:
     ChamberRectangularPartDescription x;
     ...
     Rectangle rect(x.ind1*x.grid, x.ind2*x.grid);
   (See the definition of operator* between index and grid in point.hh)
*/

// Author V.Balagura (balagura@mail.desy.de)   (21.10.98).

#ifndef setup_algorithms_hh
#define setup_algorithms_hh

#include <vector>
#include <set>
#include "ptset/point.hh"

struct ChamberRectangularPartDescription {
  Index ind1, ind2;
  Grid grid;
  ChamberRectangularPartDescription() {}
  ChamberRectangularPartDescription(const Index& i1, const Index& i2,
				    const Grid& g)
    : ind1(i1), ind2(i2), grid(g) {}
};

typedef vector<ChamberRectangularPartDescription>
Vector_Of_ChamberRectangularPartDescriptions;

/* This typedef is done to get around the bug in gcc version 2.7.2.3.f.1.
   If one writes directly
     template <class AnyChamber>
     void make_chamber_from_rectangles
     (const AnyChamber& chamber,
     vector<ChamberRectangularPartDescriptions>* rectangles) {...}
   he gets with this gcc version an error:
   "parse error before character 0254" (the message with no meaning).
   After localizing the reason why this error occured I found that
   this typedef solves the problem
*/

template <class AnyChamber>
void make_chamber_from_rectangles
(AnyChamber& chamber,
 Vector_Of_ChamberRectangularPartDescriptions* rectangles) {
  typename AnyChamber::ISection is;
  typename AnyChamber::ICell ic;
  Index ind, ind1, ind2;
  set<Index, less<Index> > left; // stores Indexes of the cells not already
  // folded into the rectangles
  set<Index, less<Index> >::iterator i_left;

  for (is=chamber.begin(); is!=chamber.end(); ++is) {
    typename AnyChamber::Section& sect = *is;
    Grid& grid = sect.grid;
    for (ic=sect.begin(); ic!=sect.end(); ++ic) left.insert((*ic).first);
    // in the beginning left contains all indexes in the section
    while ( left.size() != 0 ) { // run until all section is covered
      // by rectangles
      ind1 = *left.begin();
      for ( ind=*left.begin(); (i_left=left.find(ind)) != left.end();
	    ++ind.ix ) left.erase(i_left);
      // move horizontally to the right till the boundary
      ind2.ix = ind.ix;

      bool row_exists;
      while (1) {  // move vertically to the top untill the part of the row
	// between ind1.ix <= ix < ind2.ix is inside the chamber
	row_exists = true;
	++ind.iy;
	vector<set<Index, less<Index> >::iterator> v_row;
	for (ind.ix=ind1.ix; ind.ix!=ind2.ix; ++ind.ix) {
	  if ((i_left=left.find(ind)) == left.end() ) {
	    row_exists = false;
	    break;
	  }
	  v_row.push_back(i_left);  // mark those elements for future erasing
	}
	if ( row_exists )
	  for (unsigned int i=0; i<v_row.size(); ++i) left.erase(v_row[i]);
	else break;
      }
      ind2.iy = ind.iy;

      rectangles->push_back(ChamberRectangularPartDescription
			    (ind1, ind2, grid));
    }
  }
}
  
#endif
