////////////////////////////////////////////////////////////
// 
// Save RICH likelihood information from RTRA to RTRI,
// copy saved information back to RTRI:
//     algorithm 0 = RISE
//     algorithm 1 = RITER/RIRE
//
// copyRTRAtoRTRI(int algorithm)
// copyRTRItoRTRA(int algorithm)
//
//  returns 0 if OK
//
//  Created: 15 July, 2000 D.Dujmic
//
//  modified:
//    06 Aug 2000  (DD) maximizeRTRA - chooses higher likelihood 
//                      of two two methods
//    06 Aug 2000  (DD) favorRTRA - choses one method, if not filled
//                      takes the other method
//
//
////////////////////////////////////////////////////////////

#include "arte/arte.hh"

unsigned int nrtra;

int copyRTRAtoRTRI(int algorithm) {

  // -- setup RTRI for a new event, create a parallel table to RTRA
  static int old_event=0;
  if (old_event!=ArtePointer<EVHD>(1)->evt) {
    ArteTable<RTRI>::clear();
    nrtra=ArteTable<RTRA>::size();
    RTRI rtri;
    memset(&rtri,0,sizeof(RTRI));
    for (unsigned int i=1; i<=nrtra; i++) ArteTable<RTRI>::push_back(rtri);
    old_event=ArtePointer<EVHD>(1)->evt;
  }
  if (!nrtra) return 0;
  if (nrtra!=ArteTable<RTRI>::size()) { 
    cout<<"copyRTRAtoRTRI: nrtra!=ArteTable<RTRI>::size() !!"<<endl;
    return -1;
  }

  // -- copy RTRA for an algorithm
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    switch (algorithm) {
    case 0:
      rtri->lte =rtra->lre;
      rtri->ltmu=rtra->lrmu;
      rtri->ltpi=rtra->lrpi;
      rtri->ltk =rtra->lrk;
      rtri->ltp =rtra->lrp;
      rtri->flt =rtra->hri;
      break;
    case 1:
      rtri->lse =rtra->lre;
      rtri->lsmu=rtra->lrmu;
      rtri->lspi=rtra->lrpi;
      rtri->lsk =rtra->lrk;
      rtri->lsp =rtra->lrp;
      rtri->fls =rtra->hri;
      break;
    default :
      cout<<"copyRTRAtoRTRI: wrong algorithm specified: "<<algorithm
	  <<"  must be 0 or 1"<<endl;
      return -1;
      break;
    }
  }

  return 0;
}

int copyRTRItoRTRA(int algorithm) {

  nrtra=ArteTable<RTRI>::size();
  if (!nrtra) return 0;

  // -- copy RTRI to RTRA for an algorithm
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    switch (algorithm) {
    case 0:
      rtra->lre =rtri->lte;
      rtra->lrmu=rtri->ltmu;
      rtra->lrpi=rtri->ltpi;
      rtra->lrk =rtri->ltk;
      rtra->lrp =rtri->ltp;
      rtra->hri =rtri->flt;
      break;
    case 1:
      rtra->lre =rtri->lse;
      rtra->lrmu=rtri->lsmu;
      rtra->lrpi=rtri->lspi;
      rtra->lrk =rtri->lsk;
      rtra->lrp =rtri->lsp;
      rtra->hri =rtri->fls;
      break;
    default :
      cout<<"copyRTRItoRTRA: wrong algorithm specified: "<<algorithm
	  <<"  must be 0 or 1"<<endl;
      return -1;
      break;
    }
  }

  return 0;
}



int maximizeRTRA() {
  nrtra=ArteTable<RTRI>::size();
  if (!nrtra) return 0;
  // -- take higher likelihood of two algorithms
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    rtra->lre  = (rtri->lte >rtri->lse ) ? rtri->lte  : rtri->lse;
    rtra->lrmu = (rtri->ltmu>rtri->lsmu) ? rtri->ltmu : rtri->lsmu;
    rtra->lrpi = (rtri->ltpi>rtri->lspi) ? rtri->ltpi : rtri->lspi;
    rtra->lrk  = (rtri->ltk >rtri->lsk ) ? rtri->ltk  : rtri->lsk;
    rtra->lrp  = (rtri->ltp >rtri->lsp ) ? rtri->ltp  : rtri->lsp;
    rtra->hri  = (rtri->flt >rtri->fls ) ? rtri->flt  : rtri->fls;
  }
  // should I normalize?
  return 0;
}



int favorRTRA(int mode) {
  nrtra=ArteTable<RTRI>::size();
  if (!nrtra) return 0;
  // -- favor one of algorithms, if not filled use the other one
  // -- RISE=0, RITER/RIRE=1
  for (unsigned int i=1; i<=nrtra; i++) {
    ArtePointer<RTRI> rtri(i);
    ArtePointer<RTRA> rtra(i);
    switch (mode) {
    case 0:
      rtra->lre  = (rtri->lte >0) ? rtri->lte  : rtri->lse;
      rtra->lrmu = (rtri->ltmu>0) ? rtri->ltmu : rtri->lsmu;
      rtra->lrpi = (rtri->ltpi>0) ? rtri->ltpi : rtri->lspi;
      rtra->lrk  = (rtri->ltk >0) ? rtri->ltk  : rtri->lsk;
      rtra->lrp  = (rtri->ltp >0) ? rtri->ltp  : rtri->lsp;
      rtra->hri  = (rtri->flt >0) ? rtri->flt  : rtri->fls;
      break;
    case 1:
      rtra->lre  = (rtri->lse >0) ? rtri->lse  : rtri->lte;
      rtra->lrmu = (rtri->lsmu>0) ? rtri->lsmu : rtri->ltmu;
      rtra->lrpi = (rtri->lspi>0) ? rtri->lspi : rtri->ltpi;
      rtra->lrk  = (rtri->lsk >0) ? rtri->lsk  : rtri->ltk;
      rtra->lrp  = (rtri->lsp >0) ? rtri->lsp  : rtri->ltp;
      rtra->hri  = (rtri->fls >0) ? rtri->fls  : rtri->flt;      
      break;
    defaut:
      cout<<"favorRTRA: wrong algorithm specified: "<<mode
	  <<"  must be 0 or 1"<<endl;
      return -1;
      break;
    }
  }
  // should I normalize?
  return 0;
}
