#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "l2field.h"
#include "l2field_table.h"

#define SQ(x) ((x)*(x))

/***********
 *
 *  FieldGetValue
 *  Returns values from field parametrization lookup tables, necessary to 
 *  predict difference in slope and displacement of track when traversing
 *  magnet.
 *
 *  Input:   tx   : slope before magnet zx-plane
 *           ty   : slope before magnet zy-plane
 *           pinv : Q/p  (used to estimate higher order effects).
 *
 *           Lowest order results are independent of Q/p.
 *           If Q/p unknown by caller, set to 0. and higher order effects
 *           will be ignored.
 *
 *  Output:  Field integrals:
 *           h[0] : p/Q * Change in slope zx-plane
 *           h[1] : p/Q * Change in x-position at center of magnet
 *           h[2] : p/Q * Change in slope zy-plane
 *           h[3] : p/Q * Change in y-position at center of magnet
 *
 *           Everywhere: Change = (after - before)
 *
 *           Multiply h-vector by Q/p in calling routine to get changes
 *           in slope/position.
 *
 *  Authors: Stefania Xella, Mogens Dam
 *
 *  Changes: 
 *  04 Aug 2000 (BP) Renamed and added to Arte
 *
 ************/
void FieldGetValue ( float tx, float ty, float pinv, float h[4] ) {
  float xcell = 2.*TXLIMIT/NCELLX;       // Cell size x
  float ycell = 2.*TYLIMIT/NCELLY;       // Cell size y
  float txlow = (1./NCELLX-1.)*TXLIMIT;  // Center position of lowest cell x
  float tylow = (1./NCELLY-1.)*TYLIMIT;  //                                y
  float x, y, wxm, wym, wxp, wyp;
  float wmm, wmp, wpm, wpp;
  int ixp, iyp, ixm, iym;

  /* Find cell number */
  x   = (tx-txlow)/xcell;
  y   = (ty-tylow)/ycell;
  ixp = (int)(1.+x);        // x above
  iyp = (int)(1.+y);        // y above
  ixm = ixp-1;              // x below
  iym = iyp-1;              // y below

  /* We have defined corners of cell.      *
   * Find relative position inside cell.   *
   * Units = cell widths                   */
  wxp = x - (float)ixm;      wxm = 1. - wxp;
  wyp = y - (float)iym;      wym = 1. - wyp;

  /* Just make sure we are inside grid */
  if ( ixp < 0 ) ixp = 0;    if ( ixp >= NCELLX ) ixp = NCELLX-1;
  if ( iyp < 0 ) iyp = 0;    if ( iyp >= NCELLY ) iyp = NCELLY-1;
  if ( ixm < 0 ) ixm = 0;    if ( ixm >= NCELLX ) ixm = NCELLX-1;
  if ( iym < 0 ) iym = 0;    if ( iym >= NCELLY ) iym = NCELLY-1;

  /* Give weight of surrounding points according to relative position */
  wmm = wxm * wym;
  wmp = wxm * wyp;
  wpm = wxp * wym;
  wpp = wxp * wyp;

  h[0] = field[ixm][iym][0]*wmm + field[ixp][iym][0]*wpm +
         field[ixm][iyp][0]*wmp + field[ixp][iyp][0]*wpp;

  h[1] = field[ixm][iym][1]*wmm + field[ixp][iym][1]*wpm +
         field[ixm][iyp][1]*wmp + field[ixp][iyp][1]*wpp;

  h[2] = field[ixm][iym][2]*wmm + field[ixp][iym][2]*wpm +
         field[ixm][iyp][2]*wmp + field[ixp][iyp][2]*wpp;

  h[3] = field[ixm][iym][3]*wmm + field[ixp][iym][3]*wpm +
         field[ixm][iyp][3]*wmp + field[ixp][iyp][3]*wpp;

  /* Add parametrized second order terms */
  if ( pinv != 0. ) {
    float tx2, ty2;
    tx2 = SQ(tx);
    ty2 = SQ(ty);
    h[0] += pinv * 20.*tx*ty2*(1.+10.*tx2);
    h[1] += pinv * 31*tx;
    h[2] += pinv * ty*(-0.20+0.*tx2+800.*tx2*tx2-10000.*ty2*tx2*tx2);
    h[3] += pinv * 35.*ty*(1.+ 27.*ty2 - 36.*tx2 + 575.*tx2*ty2);
  }
}
