
// *****************************************************************************
//
// source:    
//
// type:      source code
//
// created:   26 Jul 2000
//	
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// changes:
// 26 Jul 2000 (AG) creation (code from Denis)
// 31 Jul 2000 (DD) scan for primary vertex/wire added
// 04 Aug 2000 (DD) vtchi not called before vtfit (T.Glebe)
// 16 Aug 2000 (DD) option for v0 constraint
// 17 Aug 2000 (DD) option for exclusion of v0 tracks
//
// *****************************************************************************

#include "reco/rcXvertex.h"
#include "reco/rcXsegment.h"
#include "reco/rcXmessage.h"

vertexmass::vertexmass() {
  track_list = new vector<TRACK>;
  wire_list  = new vector<ArtePointer<GTAR> >;
  findUsedWires();
}


void vertexmass::bookRVER() {
  // -- book rver
  RVER rver;
  rver.xv=v[0];
  rver.yv=v[1];
  rver.zv=v[2];
  for (int i=0; i<3; i++)
    for (int j=i; j<3; j++) rver.cv[segment::covij[i][j]]=a[i][j];
  rver.mass=Mass;
  rver.merr=Merr;
  rver.nt=track_list->size();
  ArteTable<RVER>::push_back(rver);
  ArtePointer<RVER> prver(ArteTable<RVER>::size());

  // -- book rtra-rrel-rver
  for (unsigned int i=0; i<track_list->size(); i++) {
    RREL rrel;
    rrel.flag = 1;
    rrel.chi  = chi2;
    ArteTable<RREL>::push_back(rrel);
    ArtePointer<RREL> prrel(ArteTable<RREL>::size());
    createRelation(                 prver, prrel, &RVER::rrel );
    createRelation( (*track_list)[i].rtra, prrel, &RTRA::rrel );
  }
}



int vertexmass::fitVertex() {
  if (track_list->size()<2)                  return RCX_VERTEX_CUT;
  switch (getVertAttrib().rlevel) {
  case 1: vtchi(); break;
  case 2:
  case 3: vtfit(); break;
  };
  if (vattrib.verb) printVertex();
  if (chi2>vattrib.max_vertex_chi2)          return RCX_VERTEX_CUT;
  float fpar[]={v[0],v[1],v[2], par[0],par[1]};
  // -- old approach: find wire, primary vertex
  //if (vattrib.check_wire && checkWire(fpar)) return RCX_VERTEX_CUT;
  //if (findPrimaryVertex());//                   return RCX_VERTEX_CUT;
  // -- new approach: find primary vertex, wire
  if (checkPrimaryVertex(fpar) && vattrib.check_wire) return RCX_VERTEX_CUT;
  return RCX_OK;
}


void vertexmass::printVertex() {
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    cout<<"track "<<(*it).rtra
	<<"  x="<<(*it).x[0]<<","<<(*it).x[1]<<","<<(*it).x[2]
	<<"  n="<<(*it).n[0]<<","<<(*it).n[1]<<","<<(*it).n[2]
	<<endl;
  }
  cout<<"vertex: "<<"x="<<v[0]<<","<<v[1]<<","<<v[2]
      <<"  chi="<<chi2<<endl<<endl;
}

void vertexmass::addVertexTrack(ArtePointer<RTRA> tp, ArtePointer<TOPP> to) {
  TRACK tr;
  tr.rtra=tp;
  tr.x[0]=tp->xf;
  tr.x[1]=tp->yf;
  tr.x[2]=tp->zf;
  tr.n[0]=tp->txf; tr.n[1]=tp->tyf;
  float inorm=1/sqrt(tr.n[0]*tr.n[0]+tr.n[1]*tr.n[1]+1);
  tr.n[0]=tp->txf*inorm; tr.n[1]=tp->tyf*inorm; tr.n[2]=inorm;
  tr.topp=to;
  track_list->push_back(tr);
}


void vertexmass::removeLastVertexTrack() {
  if (!track_list->size()) return;
  vector<TRACK>::iterator it=track_list->end();
  track_list->erase(--it); 
}

vector<ArtePointer<GTAR> >::iterator vertexmass::findGTAR(float *fpar) {
  // fpar[5]: x,y,z,tx,ty
  float min_drho2=vattrib.max_drho_from_wire*vattrib.max_drho_from_wire;
  vector<ArtePointer<GTAR> >::iterator iwire=wire_list->end();
  for (vector<ArtePointer<GTAR> >::iterator igtar=wire_list->begin();
       igtar!=wire_list->end(); ++igtar) {
    // -- decay behind/at a wire
    if ((*igtar)->z-fpar[2]>.5) continue; /////////// hard-coded vertex resolution!!!!
    float dz=(*igtar)->z-fpar[2];
    float dx=fpar[0]+fpar[3]*dz - (*igtar)->x;
    float dy=fpar[1]+fpar[4]*dz - (*igtar)->y;
    float drho2=dx*dx+dy*dy;
    if (drho2<min_drho2) { min_drho2=drho2; iwire=igtar; }
  }
  //cout<<"best drho2="<<min_drho2<<endl;
  return iwire;
}

ArteTable<RVER>::iterator vertexmass::findRVER(float *fpar) {
  // fpar[5]: x,y,z,tx,ty
  float min_drho2=vattrib.max_drho_from_wire*vattrib.max_drho_from_wire;
  ArteTable<RVER>::iterator iv0=ArteTable<RVER>::end();
  for (ArteTable<RVER>::iterator irver=ArteTable<RVER>::begin();
       irver!=ArteTable<RVER>::end(); ++irver) {
    // -- decay before v0
    if ((*irver)->zv-fpar[2]>.5)            continue; /////////// hard-coded vertex resolution!!!!
    if ((*irver)->nt<vattrib.min_v0_tracks) continue; 
    if ((*irver)->wire<1||(*irver)->wire>8) continue;  
    float dz=(*irver)->zv-fpar[2];
    float dx=fpar[0]+fpar[3]*dz - (*irver)->xv;
    float dy=fpar[1]+fpar[4]*dz - (*irver)->yv;
    float drho2=dx*dx+dy*dy;
    if (drho2<min_drho2) { min_drho2=drho2; iv0=(*irver); }
  }
  return iv0;
}

int vertexmass::checkPrimaryVertex(float *fpar) {
  ArteTable<RVER>::iterator irver=findRVER(fpar);
  if (irver==ArteTable<RVER>::end())                  return RCX_WIRE_CUT;
  v0_rver=(*irver);
  decay_wire=ArtePointer<GTAR>(v0_rver->wire);
  if (vattrib.verb) 
    cout<<"primary vertex: "<<v0_rver<<" "<<v0_rver->xv<<" "<<v0_rver->yv<<" "<<v0_rver->zv<<"  wire="<<v0_rver->wire<<endl;
  // -- check decay length
  if (fpar[2]-v0_rver->zv<vattrib.min_dz_from_wire)   return RCX_VERTEX_CUT;
  // -- check if tracks come from primary vertex
  if (vattrib.exclude_v0_tracks) {
    for (ArteRelationIterator<RREL> irrel=v0_rver->rrel.begin();
	 irrel!=v0_rver->rrel.end(); ++irrel) {
      ArtePointer<RTRA> v0_prong=(*irrel)->rtra;
      for (vector<TRACK>::iterator it=track_list->begin();
	   it!=track_list->end(); ++it) {
	if ((*it).rtra==v0_prong)                     return RCX_EXCLUDE_V0;
      }
    }
  }
  return RCX_OK;
}


int vertexmass::checkWire(float *fpar) {
  // -- try to find the best wire in drho
  vector<ArtePointer<GTAR> >::iterator iwire=findGTAR(fpar);
  if (iwire==wire_list->end())                      return RCX_VERTEX_CUT;
  decay_wire=(*iwire);
  if (vattrib.verb>2) {
    float dz=fpar[2]-(*iwire)->z;
    float dx=fpar[0]+fpar[3]*dz - (*iwire)->x;
    float dy=fpar[1]+fpar[4]*dz - (*iwire)->y;
    float drho2=dx*dx+dy*dy;
    cout<<"input param ="<<fpar[0]<<" "<<fpar[1]<<" "<<fpar[2]<<" "<<fpar[3]<<" "<<fpar[4]<<endl;
    cout<<"nearest wire="<<(*iwire)<<"  dz="<<dz<<"  drho="<<sqrt(drho2)<<endl;
    cout<<"             "<<(*iwire)->x<<" "<<(*iwire)->y<<" "<<(*iwire)->z<<endl;
  }
  // -- check decay length
  if (fpar[2]-(*iwire)->z<vattrib.min_dz_from_wire)   return RCX_VERTEX_CUT;
  return RCX_OK;
}


void vertexmass::findUsedWires() {
  float maxBeamTargetDx=1, maxBeamTargetDy=1;
  for (ArteTable<GTAR>::iterator igtar=ArteTable<GTAR>::begin();
       igtar!=ArteTable<GTAR>::end(); ++igtar) {
    //cout<<(*igtar)<<" "<<(*igtar)->x<<" "<<(*igtar)->y<<" "<<(*igtar)->z<<endl;
    if ( fabs((*igtar)->x)<maxBeamTargetDx &&
         fabs((*igtar)->y)<maxBeamTargetDy) { 
      wire_list->push_back((*igtar));
      if (vattrib.verb) cout<<"using wire="<<(*igtar)<<"  "
			    <<(*igtar)->x<<" "<<(*igtar)->y<<" "<<(*igtar)->z<<endl;
    }
  }
  if (!wire_list->size()) cout<<"vertexmass: cannot find target wires"<<endl;
}


int vertexmass::findPrimaryVertex() {
  int   ret=RCX_VERTEX_CUT;
  float min_dz=.3;
  float min_drho2=.11;
  for (ArteTable<RVER>::iterator irver=ArteTable<RVER>::begin();
       irver!=ArteTable<RVER>::end(); ++irver) {
    if ((*irver)->nt<vattrib.min_v0_tracks)  continue;
    if ((*irver)->wire!=decay_wire.index())  continue; 
    float dz=fabs(decay_wire->z-(*irver)->zv);
    if (dz>min_dz)                           continue; 
    float dx=(decay_wire->x-(*irver)->xv);
    float dy=(decay_wire->y-(*irver)->yv);
    float drho2=dx*dx+dy*dy;
    if (drho2>min_drho2) continue;
    min_drho2=drho2;
    v0_rver=(*irver);
    ret=RCX_OK;
  }
  return ret;
}


void vertexmass::findVertexMC() {
  if (!ArteTable<MTRA>::size()) return; 

  // -- find primary vertex
  ArtePointer<MTRA> bmtra = decay_mtra;
  v0_mver                 = bmtra->bver;
  int               flag  = v0_mver->flag;
  while (v0_mver->flag!=1) {
    bmtra=bmtra->btra;
    v0_mver=bmtra->bver;
  }

  // -- find MC vertex
  if (decay_mtra->ever.begin()==decay_mtra->ever.end()) decay_ever=decay_mtra->bver;
  else for (ArteRelationIterator<MVER> iever=decay_mtra->ever.begin(); 
	    iever!=decay_mtra->ever.end(); ++iever) decay_ever=(*iever);

  if (vattrib.verb) cout<<" primary vertex "<<v0_mver<<": "<<v0_mver->x<<" "<<v0_mver->y<<" "<<v0_mver->z<<endl;
  if (vattrib.verb) cout<<" mc vertex "<<decay_ever<<": "<<decay_ever->x<<" "<<decay_ever->y<<" "<<decay_ever->z<<endl;
}


//////////////////////////////// vertex fit 1 ////////////////////////////////
// -- minimizes chi2 deviations of tracks from unknown vertex
// -- problem can be solved algebraically, w/o fitting
// -- based on code by P.Schlein
extern "C" void rinv_(int &ndim, float *a, int &n, float *s, int &iflag);
void vertexmass::vtchi() {

  static float d[3];
  memset(&a[0][0],0x0,sizeof(float)*9);
  memset(&d[0],0x0,sizeof(float)*3);
  memset(&v[0],0x0,sizeof(float)*3);
  chi2=1e5;

  // -- loop over tracks
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    float xdotn=0;
    for (int i=0; i<3; i++)  xdotn+=(*it).x[i]*(*it).n[i];
    for (int i=0; i<3; i++) {
      a[i][i]+=1-(*it).n[i]*(*it).n[i];
      d[i]   +=(*it).x[i]-xdotn*(*it).n[i];
      for (int j=i+1; j<3; j++) {
	a[i][j]+= -(*it).n[i]*(*it).n[j];
	a[j][i]=a[i][j];
      }
    }
  }
  
  // -- compute vertex position
  int   iflag=0, ndim=3;
  float sc[3];
  rinv_(ndim, &a[0][0], ndim, sc, iflag);
  if (iflag) { return; }
  for (int i=0; i<3; i++) {
    v[i]=0;
    for (int j=0; j<3; j++) v[i]+=a[i][j]*d[j];
  }
  
  // -- compute chi^2
  chi2=0;
  float del[3] , isg[3];
  memset(&del[0],0x0,sizeof(float)*3);
  memset(&par[0],0x0,sizeof(float)*5);
  isg[0]=isg[1]=1/.0070; isg[2]=1/.0500;
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    float xvdotn=0;
    memset(&del[0],0x0,3*sizeof(float));
    for (int j=0; j<3; j++) xvdotn += ((*it).x[j]-v[j])*(*it).n[j];
    for (int j=0; j<3; j++) del[j]  = ( xvdotn*(*it).n[j] - ((*it).x[j]-v[j]) )*isg[j];
    for (int j=0; j<3; j++) chi2   += ( del[j]*del[j] );
    float pf=fabs((*it).rtra->pf);
    if (pf>1e-4) pf=1/pf;
    for (int i=0; i<3; i++) par[i] +=(*it).n[i]*pf; 
  }
  float pz=sqrt(par[0]*par[0]+par[1]*par[1]+par[2]*par[2]);
  par[0]/=pz; par[1]/=pz; par[4]=par[2];

  chi2/=(track_list->size()*2-3);
}



//////////////////////////////// vertex fit 2 ////////////////////////////////
// -- Kalman filtering method for vertexing 
// -- based on v2fit by T.Lohse
extern "C" {
  void  vtzero_();
  void  vtrini_(int &i);
  void  vtztr_(int &it1, int &it2, double &z1, double &z2);
  void  vtinit_(double *x, double *a, int &iv, int &iflag);
  void  vtadd_(int &iv, int &i, int &iflag);
  void  vtfit_(int &iv, int &momflg, int &iflag);
  float vtchi2_(int &iv, int &k, double *xk, double *qk, double &b13, 
		double &a33, double &a34, int &momflg);
  void  vtsum_(int &iv, double *par, double *cov, int &iflag);
  void  vtaddc_(int &iv, double &xM, bool *mask, int &iflag);
  void  vtmass_(int &iv, double *par, double *cov, double &chi2m, int &iflag);
}
#define NVTMAX 50
#define NTRK   100
#define NTRTOT 300
#define NCOMAX 5
extern struct {
  double xv[(NTRK+1)*NVTMAX*3], 
    xvs[NVTMAX*3], 
    qv[NTRK*NVTMAX*3], 
    qvs[NTRK*NVTMAX*3], 
    c[(NTRK+1)*NVTMAX*3*3], 
    cinv[(NTRK+1)*NVTMAX*3*3], 
    f[NTRK*NVTMAX*3*3], 
    gb[NTRK*NVTMAX*5*5], 
    w[NTRK*NVTMAX*3*3], 
    cs[3*NVTMAX*3], 
    ds[3*3*NVTMAX*NTRK], 
    es[NTRK*NVTMAX*3*3], 
    chi2[NTRK*NVTMAX], 
    chi2s[NTRK*NVTMAX],
    xnk[NTRK*NVTMAX*3],
    cnkinv[NTRK*NVTMAX*3*3], 
    cnk[NTRK*NVTMAX*3*3], 
    Ak33[NTRK*NVTMAX], 
    Ak34[NTRK*NVTMAX], 
    Bk13[NTRK*NVTMAX];
  int    fit[NVTMAX], 
    fitm[NVTMAX], 
    smooth[NVTMAX], 
    dist[NVTMAX],
    invflg[NTRTOT];
} vtkalm_;
extern struct {
  double p[NTRTOT*5], v[5*NTRTOT*5], g[5*NTRTOT*5], xmass[NTRTOT];
  int    trmult;
} vttrbf_; 
extern struct {
  int    nconst[NVTMAX];
  double xmc[NVTMAX*NCOMAX];
  bool   mask[NVTMAX*NCOMAX*NTRK];
} vtcons_;
void vertexmass::vtfit() {

  if (track_list->size()<2) return;

  // -- initialize and fill track buffer
  //double cov[3][3]; memcpy(&cov[0][0],&a[0][0],sizeof(float)*9);
  static double cov[3][3]={{1e5,0,0},{0,1e5,0},{0,0,1e5}};
  //memset(&v[0],0x0,sizeof(float)*3);
  chi2=1e5;
  static int old_event=0;
  if (vattrib.remove_old_vertices||old_event!=ArtePointer<EVHD>(1)->evt) { 
    vtzero_(); 
    old_event=ArtePointer<EVHD>(1)->evt; 
  }
  int n0=vttrbf_.trmult, ni=0;
  for (vector<TRACK>::iterator it=track_list->begin();
       it!=track_list->end(); ++it) {
    int itr=(*it).rtra.index();
    if (vattrib.verb) cout<<"vtrini itr="<<itr<<endl;
    vtrini_(itr);
    vttrbf_.xmass[n0+ni]=(*track_list)[ni].topp->mass;
    ni++;
  }
  int n1=n0+ni;
  if (vattrib.verb) {
    cout<<"vertexmass: new tracks="<<ni<<endl;
    for (int i=n0; i<n1; i++) {
      cout<<"p= "<<vttrbf_.p[i*5]<<" "<<vttrbf_.p[i*5+1]<<" "<<vttrbf_.p[i*5+2]<<" "<<vttrbf_.p[i*5+3]<<" "<<vttrbf_.p[i*5+4];
      cout<<"  m="<<vttrbf_.xmass[i]<<endl;
    }
  }


  // -- get rough vertex position
  double x[3]={v[0],v[1],v[2]};
  //if (vattrib.verb) cout<<"vtztr: "<<x[0]<<" "<<x[1]<<" "<<x[2]<<endl;

  // -- initialize fit, add tracks to vertex
  int iv, iflag=0;
  vtinit_(x, &cov[0][0], iv, iflag);
  if (vattrib.verb) cout<<"vtinit: iv="<<iv<<"  flag="<<iflag<<endl;
  if (iflag<0) return;
  for (int i=n0+1; i<=n1; i++) {
    vtadd_(iv, i, iflag);
    if (vattrib.verb) cout<<"vtadd flag="<<iflag<<endl; 
    if (iflag) return;
  }


  switch (getVertAttrib().rlevel) {
  case 2:
    // -- kalman fit w/o mass constraint
    vtfit_(iv, vattrib.momentum_flag, iflag);
    if (vattrib.verb) cout<<"vtfit flag="<<iflag<<endl;
    if (iflag) return;
    
    // -- get results
    for (int i=0; i<3; i++) v[i] = float(vtkalm_.xvs[(iv-1)*3+i]);
    chi2 = float(vtkalm_.chi2[(iv-1)+NVTMAX*(track_list->size()-1)]);
    chi2/= (2*track_list->size()-3);
    break;
    
  case 3:
    // -- kalman fit with mass constraint
    double xM=vattrib.topp->mass;
    bool   mask[NTRK];
    for (unsigned int i=0; i<track_list->size(); i++) mask[i]=true;
    vtaddc_(iv, xM, mask, iflag);
    if (vattrib.verb) cout<<"mass constraint="<<vtcons_.xmc[iv-1]<<endl;
    if (vattrib.verb) cout<<"vtmassc flag="<<iflag<<endl;
    double mPar[NTRK*3+3];
    double mCov[(NTRK*3+3)*(NTRK*3+3)];
    double mChi2=1e5;
    vtmass_(iv, mPar, mCov, mChi2, iflag);
    if (vattrib.verb) cout<<"vtmass flag="<<iflag<<endl;
    if (iflag) return;
    // -- get results
    for (int i=0; i<3; i++) v[i]=mPar[i];
    chi2=float(mChi2);
    // -- update daughter parameters
    for (unsigned int i=0; i<track_list->size(); i++) {
      ArtePointer<RTRA> rtra=(*track_list)[i].rtra;
      if (vattrib.verb)  cout<<"old pars: "<<rtra->xf<<" "<<rtra->yf<<" "<<rtra->zf<<" "
			     <<rtra->txf<<" "<<rtra->tyf<<" "<<rtra->pf<<endl;
      rtra->xf=v[0];
      rtra->yf=v[1];
      rtra->zf=v[2];
      rtra->txf=mPar[i*3+3];
      rtra->tyf=mPar[i*3+4];
      rtra->pf =((*track_list)[i].topp->chrg<0) ? -1/mPar[i*3+5] : 1/mPar[i*3+5];
      if (vattrib.verb)  cout<<"new pars: "<<rtra->xf<<" "<<rtra->yf<<" "<<rtra->zf<<" "
			     <<rtra->txf<<" "<<rtra->tyf<<" "<<rtra->pf<<endl;
    }
    break;
  }

  // -- get covariance matrix
  double dCov[5][5];
  vtsum_(iv,par,&dCov[0][0],iflag);
  if (vattrib.verb) cout<<"vtsum: iflag="<<iflag<<endl;
  // tx,ty,x,y,p -> x,y,tx,ty,p
  for (int i=0; i<5; i++) for (int j=0; j<5; j++) {
    int ii=i,jj=j;
    if (i==0 || i==1) ii+=2;
    if (j==0 || j==1) jj+=2;
    if (i==2 || i==3) ii-=2;
    if (j==2 || j==3) jj-=2;
    Cov[segment::covij[ii][jj]]=float(dCov[i][j]);
  }
  if (vattrib.verb) {
    for (int i=0; i<5; i++) { 
      for (int j=0; j<5; j++) {
	cout<<" "<<Cov[segment::covij[i][j]]<<" ";
      }
      cout<<endl;
    }
  }
}


