// *****************************************************************************
//
// source:    L3_Silicon.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#include "L3_Silicon.hh"
#include "L3_Window.hh"
#include "L3_Parameter.hh"
#include "L3_Geometry.hh"
#ifdef L3_VERBOSE
#include <iostream>
#endif
#include <set>
#include <map>
#include <algorithm>
#include <functional>
#include <stdio.h>

#ifdef L3_SLP_FRAME
#include "vdsHITB.h"
extern "C" {
  VDS_hitb* vdsslt_fillHITB( int* nhitb);
}
#endif

#ifdef L3_ARTE_FRAME 
#include "cats/cats.h"
#endif

L3_Silicon::L3_Silicon(const L3_Silicon& a) { *this = a; }

L3_Silicon& L3_Silicon::operator = (const L3_Silicon& a) {
  TrackList = a.TrackList;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      SpacePointList[i][j] = a.SpacePointList[i][j];
      SpacePoint2List[i][j] = a.SpacePoint2List[i][j];
      SpacePoint2XList[i][j] = a.SpacePoint2XList[i][j];
      for (int k=0; k<4; k++) { //Locate Plane
	HitbList[i][j][k] = a.HitbList[i][j][k];
      }
    }
  }
  spc0=a.spc0; spc1=a.spc1; spc2=a.spc2;
  trk0=a.trk0; trk1=a.trk1; trk2=a.trk2; 
  return *this;
}

void L3_Silicon::release() {
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      SpacePointList[i][j].clear();
      SpacePoint2List[i][j].clear();
      SpacePoint2XList[i][j].clear();      
      for (int k=0; k<4; k++) { //Locate Plane
	HitbList[i][j][k].clear();
      }
    }
  }
  TrackList.clear();
}
  
int L3_Silicon::nHitbs() const {
  int Count=0;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      for (int k=0; k<4; k++) { //Locate Plane
	Count += nHitbs(i,j,k);
      }
    }
  }
  return Count;
}

int L3_Silicon::nHitbs(int sl,int sec,int lay) const {
  return HitbList[sl][sec][lay].size();
}

int L3_Silicon::nPoints() const {
  int Count=0;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      Count += (SpacePointList[i][j]).size();
    }
  }
  return Count;
}

vector<L3_Hitb> 
L3_Silicon::HList() const {
  vector<L3_Hitb> tmpList;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      for (int k=0; k<4; k++) { //Locate Plane
	for (vector<L3_Hitb>::const_iterator hit = (HitbList[i][j][k]).begin();
	     hit != (HitbList[i][j][k]).end(); ++hit) {
	  tmpList.push_back(*hit);
	}
      }
    }
  }
  return tmpList;
}

vector<L3_SpacePoint>
L3_Silicon::Points() const{
  vector<L3_SpacePoint> TmpSpList;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      for (vector<L3_SpacePoint>::const_iterator SpacePointListIt
             = (SpacePointList[i][j]).begin();
           SpacePointListIt != (SpacePointList[i][j]).end();
           ++SpacePointListIt) {
        TmpSpList.push_back(*SpacePointListIt);
      }
    }
  }
  return TmpSpList;
}

void L3_Silicon::checkCompatibility(vector<L3_Track> &t){  
#ifdef L3_VERBOSE
  if( L3_Parameter::Trace() ) cout << "checkCompatibility: START" << endl;
#endif
  float ratio=0;
  do {
    int n=1;  
    ratio = -1;
    vector<L3_Track>::iterator rem=t.end();
    for(vector<L3_Track>::iterator i=t.begin();i!=t.end();i++,n++){
      vector<L3_Hitb> v = i->HList();
      float r = 1.0/float(v.size());
      //cout << n << ") " <<  v.size() << " hits, ";
      vector<L3_Hitb> in_other_tracks;
      for(vector<L3_Track>::iterator ii=t.begin();v.size()>0 && ii!=t.end();ii++) if(i!=ii){
	vector<L3_Hitb> h = ii->HList();
	for(vector<L3_Hitb>::iterator k=h.begin();v.size()>0 && k!=h.end();k++){
	  vector<L3_Hitb>::iterator res = find(v.begin(),v.end(),*k);
	  if( res!=v.end() ){
	    in_other_tracks.push_back(*res);
	    v.erase(res);
	  }
	}
      }
      r *= float(in_other_tracks.size());
      //cout << in_other_tracks.size() << " hits in other tracks " << 100.0*r << endl;
      if( r>ratio ){
	ratio = r;
	rem = i;
      }
    }
    if( ratio >= 1.0 ) t.erase(rem);          
  } while(ratio>=1.0 && t.size()>0);
#ifdef L3_VERBOSE
  trk2 = t.size();
  if( L3_Parameter::Trace() ) cout << "checkCompatibility: STOP " << trk2 << endl;
#endif
}

void L3_Silicon::reconstruct(L3_Timer& Timer) {
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::reconstruct START" << endl;
#endif
  if (L3_Parameter::run() == 98) {
    Timer.start_PointReco();
    SpacePointReco();
    SpacePairReco();
    PurgeSpacePoints();
    Timer.stop_PointReco();
    Timer.start_TrackReco();
    SortSpacePoints();
    SortSpacePairs();
    TrackReconstruction98();
    findValidTracks();
    Timer.stop_TrackReco();
  } else if( L3_Parameter::run()==0 ) {
    Timer.start_PointReco();
    SpacePointRecoChi2();
    PurgeSpacePoints();
    SpacePointAfterburner();
    Timer.stop_PointReco();
    Timer.start_TrackReco();
    SortSpacePoints();    
    TrackReconstructionPrime();
    findValidTracks();
#if 0
    checkCompatibility(TrackList);
#endif
    Timer.stop_TrackReco();    
    //unused();
  } else if(L3_Parameter::run()==-1) {
#ifdef L3_ARTE_FRAME
    Timer.start_PointReco();
    Timer.stop_PointReco();
    Timer.start_TrackReco();
    CatsEvent();
    fillTracksFromRseg();
    Timer.stop_TrackReco();    
#endif    
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::reconstruct STOP" << endl;
#endif
}

void L3_Silicon::SpacePointReco() {
//Reconstruct SpacePoints
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePointReco START" << endl;
#endif
  float dst=0.;
  float x1p=0.;
  float x3p=0.;
  float PredictHit3x=0.;
  float PredictHit4x=0.;
  float Distanz=0.0;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      (SpacePointList[i][j]).reserve(500);
      for (int P=0; P<=1; P++) {
	vector<L3_SpacePoint> SpacePairList;
	SpacePairList.reserve(500);
	int StP = L3_Geometry::getPlaneIndex(i,j,P,0);
	int P3  = L3_Geometry::getPlaneIndex(i,j,P,1);
	int P4  = L3_Geometry::getPlaneIndex(i,j,P,2);
	const L3_Plane& Plane   = L3_Geometry::getPlane(i,j,P);
	const L3_Plane& StPlane = L3_Geometry::getPlane(i,j,StP);
	const L3_Plane& Plane3  = L3_Geometry::getPlane(i,j,P3);
	const L3_Plane& Plane4  = L3_Geometry::getPlane(i,j,P4);
//  	cout << "Plane  :"  << Plane << endl;
//  	cout << "StPlane:"  << StPlane << endl;
//  	cout << "Plane3 :"  << Plane3 << endl;
//  	cout << "Plane4 :"  << Plane4 << endl;
	Distanz = ((Plane.zfw(0)+StPlane.zfw(0)) * 0.5) - L3_Geometry::ztarget();
// Anmerkung zum Stereowinkel
// fabs(Plane.cos()) ~ 1 fuer vertikale Draehte (+/- 2.5 grad zur y-Achse)
// 	    cout << " TEST : dst= " << dst << endl;
	if ( fabs( Plane.cos() ) > 0.5 ) { // +-2.5 degree layers
	  dst = fabs(2.0 * AlphaS * ( Plane.symx() - Plane.symn() ));
	  for (vector<L3_Hitb>::iterator hit = (HitbList[i][j][P]).begin();
	       hit != (HitbList[i][j][P]).end(); ++hit) {
	    x1p = Plane.icos()
	      * ( ( (Distanz / (Plane.zfw(0) - L3_Geometry::ztarget())) * hit->x() )
		  + (Plane.symn() * Plane.sin() ) );
// 	    cout << "> Hit : " << hit->hit() << "   x1p = " << x1p << endl;
	    for (vector<L3_Hitb>::iterator Sthit = (HitbList[i][j][StP]).begin();
		 Sthit!= (HitbList[i][j][StP]).end(); ++Sthit) {
	      x3p = StPlane.icos() 
		* ( (( Distanz / (StPlane.zfw(0) - L3_Geometry::ztarget())) * Sthit->x())
		    + (StPlane.symn() * StPlane.sin()) );
// 	      cout << ">>>  StHit : " 
// 		   << Sthit->hit() << "   x3p = " << x3p << endl;
	      if ( (x3p >= (x1p - dst - L3_Parameter::margin() ) ) 
		   && ( x3p <= (x1p + L3_Parameter::margin())  )) {
		//-HitsInStereo.push_back(*Sthit);
		L3_SpacePoint Paar(*hit, *Sthit);
		// cout << " Paar: " << Paar << endl;
	        SpacePairList.push_back(Paar);
	      }
	    }
	  }
	} else { // 90+-2.5 degree layers
// Anmerkung zum Stereowinkel
// fabs(Plane.cos()) ~ 0 fuer horizontale Draehte (90+/-2.5 grad zur y-Achse)
	  dst = fabs(2.0 * AlphaS * ( Plane.sxmx() - Plane.sxmn() ));
// 	    cout << " DST 2=" << dst << endl;
	  for (vector<L3_Hitb>::iterator hit = (HitbList[i][j][P]).begin();
	       hit != (HitbList[i][j][P]).end(); ++hit) {
	    x1p = Plane.isin()
	      * ( (Plane.sxmn() * Plane.cos())
		  - (hit->x() * (Distanz / (Plane.zfw(0) - L3_Geometry::ztarget()))) );
	    for (vector<L3_Hitb>::iterator Sthit = (HitbList[i][j][StP]).begin();
		 Sthit != (HitbList[i][j][StP]).end(); ++Sthit) {
	      x3p = StPlane.isin() * 
		( (StPlane.sxmn() * StPlane.cos())
		  - (Sthit->x() * (Distanz / (StPlane.zfw(0) - L3_Geometry::ztarget())) ) );
	      if ((x3p <= (x1p + dst + L3_Parameter::margin())) 
		  && (x3p >= (x1p - L3_Parameter::margin()))) {
		L3_SpacePoint Paar(*hit, *Sthit);
	        //cout << " Paar: " << Paar << endl;
	        SpacePairList.push_back(Paar);
	      }
	    }
	  }
	}
//  	cout << "TLTRECO: SpacePairList SL, Quadrant, Pair, Size: " 
//  	     << i << " "  << j << " " << P << " "
//  	     << SpacePairList.size() << " " << endl;
	for (vector<L3_SpacePoint>::iterator SpacePairListIt = SpacePairList.begin();
	     SpacePairListIt != SpacePairList.end(); ++SpacePairListIt) {
	  L3_SpacePoint& Paar = *SpacePairListIt;
	  vector<L3_Hitb> HitListLayer3;
	  vector<L3_Hitb> HitListLayer4;
//  	  cout << "TLTRECO: Debug Paar: x= " << Paar.x() 
// 	       << " y= " << Paar.y() << endl;
//  	  cout << "TLTRECO: Hitblist(Plane3).size " 
	  /// 	  << HitbList[i][j][P3].size() << endl;
	  PredictHit3x = ( (Plane3.zfw(0) - L3_Geometry::ztarget()) / Distanz) *
	    ( (Paar.x() * Plane3.cos()) -
	      (Paar.y() * Plane3.sin()) );
// 	  cout << " Predict Hit3x = " << PredictHit3x << endl;
	  for (vector<L3_Hitb>::iterator hit3= (HitbList[i][j][P3]).begin();
	       hit3 != (HitbList[i][j][P3]).end(); ++hit3) {
// 	    cout << " hit3->x() = " << hit3->x() << endl;
	    if ( fabs( PredictHit3x - hit3->x())<= L3_Parameter::StereoResol(i)) {
	      HitListLayer3.push_back(*hit3);
	    }
	  }
	  PredictHit4x = ( (Plane4.zfw(0) - L3_Geometry::ztarget()) / Distanz) * 
	    ( (Paar.x() * Plane4.cos()) -
	      (Paar.y() * Plane4.sin()) );
	  for (vector<L3_Hitb>::iterator hit4 = (HitbList[i][j][P4]).begin();
	       hit4 != (HitbList[i][j][P4]).end(); ++hit4) {
	    if ( fabs( PredictHit4x - hit4->x())<= L3_Parameter::StereoResol(i)) {
	      HitListLayer4.push_back(*hit4);
	    }
	  }
// 	  cout << "TLTRECO: # Hits in 3rd /4th  Layer "
//  	       << HitListLayer3.size() << " / "
// 	       << HitListLayer4.size() << endl;
	  if (HitListLayer3.size() > 0 ) {
	    if (HitListLayer4.size() > 0 ) {
	      for (vector<L3_Hitb>::iterator hit3= HitListLayer3.begin();
		   hit3 != HitListLayer3.end(); ++hit3) {
		for (vector<L3_Hitb>::iterator hit4 = HitListLayer4.begin();
		     hit4 != HitListLayer4.end(); ++hit4) {
		  L3_SpacePoint Paerchen(Paar);
		  Paerchen.add(*hit3, *hit4);
		  (SpacePointList[i][j]).push_back(Paerchen);
		}
	      }
	    } else {
	      for (vector<L3_Hitb>::iterator hit3= HitListLayer3.begin();
		   hit3 != HitListLayer3.end(); ++hit3) {
		L3_SpacePoint Paerchen(Paar);
		Paerchen.add(*hit3);
		(SpacePointList[i][j]).push_back(Paerchen);
	      }
	    }
	  } else if (HitListLayer4.size() > 0 ) {
	    for (vector<L3_Hitb>::iterator hit4 = HitListLayer4.begin();
		 hit4!= HitListLayer4.end(); ++hit4) {
	      L3_SpacePoint Paerchen(Paar);
	      Paerchen.add(*hit4);
	      (SpacePointList[i][j]).push_back(Paerchen);
	    }	
	  }
	}
      }
    } //sector    
  }//superlayer 
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePointReco STOP, # of points is " << nPoints() << endl;
#endif
}

void L3_Silicon::SpacePointRecoChi2() {
//Reconstruct SpacePoints
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePointReco START" << endl;
#endif
  float dst=0.;
  float x1p=0.;
  float x3p=0.;
  float PredictHit3x=0.;
  float PredictHit4x=0.;
  float Distanz=0.;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      (SpacePointList[i][j]).reserve(500);
      for (int P=0; P<=1; P++) {
	vector<L3_SpacePoint> SpacePairList;
	SpacePairList.reserve(500);
	int StP = L3_Geometry::getPlaneIndex(i,j,P,0);
	int P3  = L3_Geometry::getPlaneIndex(i,j,P,1);
	int P4  = L3_Geometry::getPlaneIndex(i,j,P,2);
	const L3_Plane& Plane   = L3_Geometry::getPlane(i,j,P);
	const L3_Plane& StPlane = L3_Geometry::getPlane(i,j,StP);
	const L3_Plane& Plane3  = L3_Geometry::getPlane(i,j,P3);
	const L3_Plane& Plane4  = L3_Geometry::getPlane(i,j,P4);
	Distanz = ((Plane.zfw(0)+StPlane.zfw(0)) * 0.5) - L3_Geometry::ztarget();
// Anmerkung zum Stereowinkel
// fabs(Plane.cos()) ~ 1 fuer vertikale Draehte (+/- 2.5 grad zur y-Achse)
// 	    cout << " TEST : dst= " << dst << endl;
	if ( fabs( Plane.cos() ) > 0.5 ) { // +-2.5 degree layers
	  dst = fabs(2.0 * AlphaS * ( Plane.symx() - Plane.symn() ));
	  for (vector<L3_Hitb>::iterator hit = (HitbList[i][j][P]).begin();
	       hit != (HitbList[i][j][P]).end(); ++hit) {
	    x1p = Plane.icos()
	      * ( ( (Distanz / (Plane.zfw(0) - L3_Geometry::ztarget())) * hit->x() )
		  + (Plane.symn() * Plane.sin() ) );
// 	    cout << "> Hit : " << hit->hit() << "   x1p = " << x1p << endl;
	    for (vector<L3_Hitb>::iterator Sthit = (HitbList[i][j][StP]).begin();
		 Sthit!= (HitbList[i][j][StP]).end(); ++Sthit) {
	      x3p = StPlane.icos() 
		* ( (( Distanz / (StPlane.zfw(0) - L3_Geometry::ztarget())) * Sthit->x())
		    + (StPlane.symn() * StPlane.sin()) );
// 	      cout << ">>>  StHit : " 
// 		   << Sthit->hit() << "   x3p = " << x3p << endl;
	      if ( (x3p >= (x1p - dst - L3_Parameter::margin() ) ) 
		   && ( x3p <= (x1p + L3_Parameter::margin())  )) {
		//-HitsInStereo.push_back(*Sthit);
		L3_SpacePoint Paar(*hit, *Sthit);
		// cout << " Paar: " << Paar << endl;
	        SpacePairList.push_back(Paar);
	      }
	    }
	  }
	} else { // 90+-2.5 degree layers
// Anmerkung zum Stereowinkel
// fabs(Plane.cos()) ~ 0 fuer horizontale Draehte (90+/-2.5 grad zur y-Achse)
	  dst = fabs(2.0 * AlphaS * ( Plane.sxmx() - Plane.sxmn() ));
// 	    cout << " DST 2=" << dst << endl;
	  for (vector<L3_Hitb>::iterator hit = (HitbList[i][j][P]).begin();
	       hit != (HitbList[i][j][P]).end(); ++hit) {
	    x1p = Plane.isin()
	      * ( (Plane.sxmn() * Plane.cos())
		  - (hit->x() * (Distanz / (Plane.zfw(0) - L3_Geometry::ztarget()))) );
	    for (vector<L3_Hitb>::iterator Sthit = (HitbList[i][j][StP]).begin();
		 Sthit != (HitbList[i][j][StP]).end(); ++Sthit) {
	      x3p = StPlane.isin() * 
		( (StPlane.sxmn() * StPlane.cos())
		  - (Sthit->x() * (Distanz / (StPlane.zfw(0) - L3_Geometry::ztarget())) ) );
	      if ((x3p <= (x1p + dst + L3_Parameter::margin())) 
		  && (x3p >= (x1p - L3_Parameter::margin()))) {
		L3_SpacePoint Paar(*hit, *Sthit);
	        //cout << " Paar: " << Paar << endl;
	        SpacePairList.push_back(Paar);
	      }
	    }
	  }
	}
//  	cout << "TLTRECO: SpacePairList SL, Quadrant, Pair, Size: " 
//  	     << i << " "  << j << " " << P << " "
//  	     << SpacePairList.size() << " " << endl;
	for (vector<L3_SpacePoint>::iterator SpacePairListIt = SpacePairList.begin();
	     SpacePairListIt != SpacePairList.end(); ++SpacePairListIt) {
	  L3_SpacePoint& Paar = *SpacePairListIt;
	  vector<L3_Hitb> HitListLayer3;
	  vector<L3_Hitb> HitListLayer4;
//  	  cout << "TLTRECO: Debug Paar: x= " << Paar.x() 
// 	       << " y= " << Paar.y() << endl;
//  	  cout << "TLTRECO: Hitblist(Plane3).size " 
	  /// 	  << HitbList[i][j][P3].size() << endl;
	  PredictHit3x = ( (Plane3.zfw(0) - L3_Geometry::ztarget()) / Distanz) *
	    ( (Paar.x() * Plane3.cos()) -
	      (Paar.y() * Plane3.sin()) );
// 	  cout << " Predict Hit3x = " << PredictHit3x << endl;
	  for (vector<L3_Hitb>::iterator hit3= (HitbList[i][j][P3]).begin();
	       hit3 != (HitbList[i][j][P3]).end(); ++hit3) {
// 	    cout << " hit3->x() = " << hit3->x() << endl;
	    if ( fabs( PredictHit3x - hit3->x())<= L3_Parameter::StereoResol(i)) {
	      HitListLayer3.push_back(*hit3);
	    }
	  }
	  PredictHit4x = ( (Plane4.zfw(0) - L3_Geometry::ztarget()) / Distanz) * 
	    ( (Paar.x() * Plane4.cos()) -
	      (Paar.y() * Plane4.sin()) );
	  for (vector<L3_Hitb>::iterator hit4 = (HitbList[i][j][P4]).begin();
	       hit4 != (HitbList[i][j][P4]).end(); ++hit4) {
	    if ( fabs( PredictHit4x - hit4->x())<= L3_Parameter::StereoResol(i)) {
	      HitListLayer4.push_back(*hit4);
	    }
	  }
// 	  cout << "TLTRECO: # Hits in 3rd /4th  Layer "
//  	       << HitListLayer3.size() << " / "
// 	       << HitListLayer4.size() << endl;
	  if (HitListLayer3.size() > 0 ) {
	    if (HitListLayer4.size() > 0 ) {
	      for (vector<L3_Hitb>::iterator hit3= HitListLayer3.begin();
		   hit3 != HitListLayer3.end(); ++hit3) {
		for (vector<L3_Hitb>::iterator hit4 = HitListLayer4.begin();
		     hit4 != HitListLayer4.end(); ++hit4) {
		  L3_SpacePoint Paerchen(Paar);
		  Paerchen.add(*hit3, *hit4);
		  if (Paerchen.chiSquare() > 75.) {
		    L3_SpacePoint Paerchen3(Paar);
		    L3_SpacePoint Paerchen4(Paar);
		    Paerchen3.add(*hit3);
		    Paerchen4.add(*hit4);
		    if (Paerchen3.chiSquare() < Paerchen4.chiSquare()) {
		      (SpacePointList[i][j]).push_back(Paerchen3);
		    } else {
		      (SpacePointList[i][j]).push_back(Paerchen4);
		    }
		  } else {
		    (SpacePointList[i][j]).push_back(Paerchen);
		  }
		}
	      }
	    } else {
	      for (vector<L3_Hitb>::iterator hit3= HitListLayer3.begin();
		   hit3 != HitListLayer3.end(); ++hit3) {
		L3_SpacePoint Paerchen(Paar);
		Paerchen.add(*hit3);
		(SpacePointList[i][j]).push_back(Paerchen);
	      }
	    }
	  } else if (HitListLayer4.size() > 0 ) {
	    for (vector<L3_Hitb>::iterator hit4 = HitListLayer4.begin();
		 hit4!= HitListLayer4.end(); ++hit4) {
	      L3_SpacePoint Paerchen(Paar);
	      Paerchen.add(*hit4);
	      SpacePointList[i][j].push_back(Paerchen);
	    }	
	  } else {  //no confirming hit found for this space pair 
	    //keep it if there are less than 4 layers in this sl,sec
	    if( L3_Geometry::nPlanes(i,j) < 4 ) SpacePointList[i][j].push_back(Paar);
	  }
	}
      }
    } //sector    
  }//superlayer 
#ifdef L3_VERBOSE
  spc0 = nPoints();
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePointReco STOP, # of points is " << spc0 << endl;
#endif
}

void show(vector<L3_Hitb> b){
  vector<int> s;
  for(vector<L3_Hitb>::iterator i=b.begin();i!=b.end();i++) s.push_back( i->hit() );
  sort(s.begin(),s.end());
  for(vector<int>::iterator i=s.begin();i!=s.end();i++) cout << "[" << *i << "] ";  
}

void L3_Silicon::PurgeSpacePoints() {  
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::PurgeSpacePoints" << endl;
#endif
  for (int i=0; i<7; i++) { //Locate SuperLayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      for (vector<L3_SpacePoint>::iterator outer = (SpacePointList[i][j]).begin();
	   outer != (SpacePointList[i][j]).end(); ) {
	int outerHits = outer->numHits();
	int Break = 0;
	for (vector<L3_SpacePoint>::iterator inner = outer+1;
	     !Break && inner!=SpacePointList[i][j].end(); ) {
	  int innerHits = inner->numHits();
	  int commonHits = outer->getCommonHits(*inner);
	  if( outerHits>=innerHits ){ 
	    if( innerHits==commonHits ){ //inner has only hits which are also present in outer
	      SpacePointList[i][j].erase(inner);
	    } else
	      inner++;
	  } else if( outerHits==commonHits ){ //outer has only hits which are also present in inner
	    *outer = *inner;
	    SpacePointList[i][j].erase(inner);
	    Break = 1;  //repeat all for this spacepoint with more hits
	  } else 
	    inner++;	      
	}
	if(!Break){
	  if( L3_Geometry::nPlanes(i,j)<4 ){
	    vector<L3_Hitb> hits;
	    outer->getCopyOfHitbs(hits);
	    for(vector<L3_Hitb>::iterator it=hits.begin();it!=hits.end();it++){
	      int lay = it->lay()-1;
	      vector<L3_Hitb>::iterator search = find(HitbList[i][j][lay].begin(),HitbList[i][j][lay].end(),*it);
	      if( search != HitbList[i][j][lay].end() )	search->setFlag(1);	      
	    }
	  }
	  outer++;
	}
      }
    }
  }
#ifdef L3_VERBOSE
  spc1 = nPoints();
  if(L3_Parameter::Trace()) cout << "L3_Silicon::PurgeSpacePoints STOP, # of points is " << spc1 << endl;
#endif
}

void L3_Silicon::PurgeSpacePointsOld() {  
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::PurgeSpacePoints" << endl;
#endif
  for (int i=0; i<7; i++) { //Locate SuperLayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      for (vector<L3_SpacePoint>::iterator outer = (SpacePointList[i][j]).begin();
	   outer != (SpacePointList[i][j]).end(); outer++) {
	for (vector<L3_SpacePoint>::iterator inner = outer+1;
	    inner != (SpacePointList[i][j]).end(); ) {
	  if( *inner == *outer ) 
	    SpacePointList[i][j].erase(inner);
	  else 
	    inner++;	    
	}
      }
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::PurgeSpacePoints STOP, # of points is " << nPoints() << endl;
#endif
}

void L3_Silicon::SpacePointAfterburner() {  
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePointAfterburner" << endl;
#endif
  for (int i=0; i<7; i++) { //Locate SuperLayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      if( L3_Geometry::nPlanes(i,j) < 4 ){
	//we stop the afterburner if the number of 2-hit- points exceeds the
	//half the number of 3-or-4-hit-points
	int n = SpacePointList[i][j].size();
	int cut = n/2;
	int count = 0;	
	for(int pl=0;pl<=1;pl++){	  
	  for(vector<L3_Hitb>::iterator k=HitbList[i][j][pl].begin();k!=HitbList[i][j][pl].end();k++) if(k->getFlag()==0){
	    for(int other=1;other<=2;other++){
	      int P34 = L3_Geometry::getPlaneIndex(i,j,pl,other);
	      for(vector<L3_Hitb>::iterator q=HitbList[i][j][P34].begin();q!=HitbList[i][j][P34].end();q++) if(q->getFlag()==0){
		L3_SpacePoint point(*k,*q,0);
		if( ++count > cut ){
		  //cout << "n=" << n << " cut=" << cut << " count=" << count;  
		  SpacePointList[i][j].erase(SpacePointList[i][j].begin()+n,SpacePointList[i][j].end());
		  //cout << " newsize=" << SpacePointList[i][j].size() << endl;
		  goto nextSuperLayerSector;
		} else
		  SpacePointList[i][j].push_back(point);
	      }
	    }	    
	  }
	}
      }
    nextSuperLayerSector:;
    }
  }
#ifdef L3_VERBOSE
  spc2 = nPoints();
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePointAfterburner STOP, # of points is " << spc2 << endl;
#endif
}

class RP_less : public binary_function< L3_SpacePoint, L3_SpacePoint, bool > {
public:
  inline
  bool operator() (const L3_SpacePoint& R , const L3_SpacePoint& S) const {
    return ( R.x() < S.x() ); 
  }
};

void L3_Silicon::SortSpacePoints() {
// Sort SpacePointList
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SortSpacePoints" << endl;
#endif
  for (int i=0; i<7; i++) { //Locate SuperLayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      sort( (SpacePointList[i][j]).begin(), (SpacePointList[i][j]).end(),
	    RP_less() );
#ifdef L3_DEBUG_SILICON
      for (vector<L3_SpacePoint>::iterator RIt 
 	     = (SpacePointList[i][j]).begin();
 	   RIt != (SpacePointList[i][j]).end(); ++RIt) {
	cout << *RIt;
      }
#endif
    }
  }
}
  
class RPWin_less : public binary_function< L3_SpacePoint, L3_Window, bool > {
public:
  inline
  bool operator() (const L3_SpacePoint& R , const L3_Window& P) const {
    return ( R.x() < P.xmin() ); 
  }
};

class RPWin_greater : public binary_function< L3_SpacePoint, L3_Window, bool > {
public:
  inline
  bool operator() (const L3_SpacePoint& R , const L3_Window& P) const {
    return ( R.x() < P.xmax() );
  }
};

void L3_Silicon::SpacePairReco() {
//Reconstruct SpacePairs
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SpacePairReco START" << endl;
#endif
  float dst=0.;
  float x1p=0.;
  float x3p=0.;
  float Distanz=0.;
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      if ((SpacePointList[i][j]).size() == 0) {
	vector<L3_SpacePoint> SpacePairList;
	SpacePairList.reserve(500);
	for (int P=0; P<=1; P++) {
	  int StP = L3_Geometry::getPlaneIndex(i,j,P,0);
	  const L3_Plane& Plane   = L3_Geometry::getPlane(i,j,P);
	  const L3_Plane& StPlane = L3_Geometry::getPlane(i,j,StP);
//          cout << "Plane  :"  << Plane << endl;
// 	    cout << "StPlane:"  << StPlane << endl;
	  Distanz = ((Plane.zfw(0)+StPlane.zfw(0)) * 0.5) - L3_Geometry::ztarget();
	  for (vector<L3_Hitb>::iterator hit = (HitbList[i][j][P]).begin();
	       hit != (HitbList[i][j][P]).end(); ++hit) {
	    vector<L3_Hitb> HitsInStereo;
// Anmerkung zum Stereowinkel
// fabs(Plane.cos()) ~ 1 fuer vertikale Draehte (+/- 2.5 grad zur y-Achse)
	    if ( fabs( Plane.cos() ) > 0.5 ) { // +-2.5 degree layers
	      dst = fabs(2.0 * AlphaS * ( Plane.symx() - Plane.symn() ));
// 	    cout << " TEST : dst= " << dst << endl;
	      x1p = Plane.icos() 
		* ( ( (Distanz / (Plane.zfw(0) - L3_Geometry::ztarget())) * hit->x() )
		    + (Plane.symn() * Plane.sin() ) );
// 	    cout << "> Hit : " << hit->hit() << "   x1p = " << x1p << endl;
	      for (vector<L3_Hitb>::iterator Sthit = (HitbList[i][j][StP]).begin();
		   Sthit!= (HitbList[i][j][StP]).end(); ++Sthit) {
		x3p = StPlane.icos() 
		  * ( (( Distanz / (StPlane.zfw(0) - L3_Geometry::ztarget())) * Sthit->x())
		      + (StPlane.symn() * StPlane.sin()) );
// 	      cout << ">>>  StHit : " 
// 		   << Sthit->hit() << "   x3p = " << x3p << endl;
		if ( (x3p >= (x1p - dst - L3_Parameter::margin() ) ) 
		     && ( x3p <= (x1p + L3_Parameter::margin())  )) {
		  HitsInStereo.push_back(*Sthit);
		}
	      }
	    } else { // 90+-2.5 degree layers
// Anmerkung zum Stereowinkel
// fabs(Plane.cos()) ~ 0 fuer horizontale Draehte (90+/-2.5 grad zur y-Achse)
	      dst = fabs(2.0 * AlphaS * ( Plane.sxmx() - Plane.sxmn() ) );
// 	    cout << " DST 2=" << dst << endl;
	      x1p = Plane.isin() 
		* ( (Plane.sxmn() * Plane.cos())
		    - (hit->x() * (Distanz / (Plane.zfw(0) - L3_Geometry::ztarget()))) );
	      for (vector<L3_Hitb>::iterator Sthit = (HitbList[i][j][StP]).begin();
		   Sthit != (HitbList[i][j][StP]).end(); ++Sthit) {
		x3p = StPlane.isin() * 
		  ( (StPlane.sxmn() * StPlane.cos())
		    - (Sthit->x() *	(Distanz / (StPlane.zfw(0) - L3_Geometry::ztarget())) ) );
		if ((x3p <= (x1p + dst + L3_Parameter::margin())) 
		    && (x3p >= (x1p - L3_Parameter::margin()))) {
		  HitsInStereo.push_back(*Sthit);
		}
	      }
	    }
	    for (vector<L3_Hitb>::iterator Sthit = HitsInStereo.begin();
		 Sthit != HitsInStereo.end(); ++Sthit) {
	      L3_SpacePoint Paar(*hit, *Sthit);
	      SpacePairList.push_back(Paar);
	    }
	  }
	}
	SpacePoint2List[i][j] = SpacePairList;
      }
      if ((SpacePoint2List[i][j]).size() == 0) {
	vector<L3_SpacePoint> SpacePairList;
	SpacePairList.reserve(500);
	for (int p=0; p<=1; p++) {
	  int P = 2*p;
	  int StP = 1 + 2*p;
	  const L3_Plane& Plane   = L3_Geometry::getPlane(i,j,P);
	  const L3_Plane& StPlane = L3_Geometry::getPlane(i,j,StP);
	  Distanz = ((Plane.zfw(0)+StPlane.zfw(0)) * 0.5) - L3_Geometry::ztarget();
	  for (vector<L3_Hitb>::iterator hit = (HitbList[i][j][P]).begin();
	       hit != (HitbList[i][j][P]).end(); ++hit) {
	    for (vector<L3_Hitb>::iterator Sthit=(HitbList[i][j][StP]).begin();
		 Sthit != (HitbList[i][j][StP]).end(); ++Sthit) {
	      L3_SpacePoint Paar(*hit, *Sthit, 1);
	      SpacePairList.push_back(Paar);
	    }
	  }
	}
	SpacePoint2XList[i][j] = SpacePairList;
      }
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << endl << "L3_Silicon::SpacePairReco STOP" << endl;
#endif
}

void L3_Silicon::SortSpacePairs() {
// Sort SpacePointList
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SortSpacePairs START" << endl;
#endif
  for (int i=0; i<7; i++) { //Locate SuperLayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      sort( (SpacePoint2List[i][j]).begin(), (SpacePoint2List[i][j]).end(),
	    RP_less() );
#ifdef L3_DEBUG_SILICON
      for (vector<L3_SpacePoint>::iterator RIt 
 	     = (SpacePoint2List[i][j]).begin();
 	   RIt != (SpacePoint2List[i][j]).end(); ++RIt) {
	cout << *RIt;
      }
#endif
      sort( (SpacePoint2XList[i][j]).begin(), (SpacePoint2XList[i][j]).end(),
	    RP_less() );
#ifdef L3_DEBUG_SILICON
      for (vector<L3_SpacePoint>::iterator RIt 
 	     = (SpacePoint2XList[i][j]).begin();
 	   RIt != (SpacePoint2XList[i][j]).end(); ++RIt) {
	cout << *RIt;
      }
#endif
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::SortSpacePairs STOP" << endl;
#endif
}

void L3_Silicon::TrackReconstruction(){
//Track reconstruction
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::TrackReconstruction START" << endl;
#endif 
  TrackList.reserve(5000);
  int Trackcounter =0;
  for (int SuperLayer=6; SuperLayer>=2; SuperLayer--) {
    for (int Sector=0; Sector<4; Sector++) {
      for (vector<L3_SpacePoint>::iterator SeedIt 
	     = (SpacePointList[SuperLayer][Sector]).begin();
	   SeedIt != (SpacePointList[SuperLayer][Sector]).end();
	   ++SeedIt) {
	L3_Track Track(Trackcounter, *SeedIt);
	Trackcounter++;
	SearchNextSL(Track,SuperLayer, 0);	
      }
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::TrackReconstruction STOP, # of tracks " << nTracks() << endl;
#endif 
}

void L3_Silicon::TrackReconstructionPrime(){
//Track reconstruction
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::TrackReconstructionPrime START" << endl;
#endif 
  TrackList.reserve(5000);
  int Trackcounter =0;
  for (int SuperLayer=6; SuperLayer>=2; SuperLayer--) {
    for (int Sector=0; Sector<4; Sector++) {
      for (vector<L3_SpacePoint>::iterator SeedIt 
	     = (SpacePointList[SuperLayer][Sector]).begin();
	   SeedIt != (SpacePointList[SuperLayer][Sector]).end();
	   ++SeedIt) {
	L3_Track Track(Trackcounter, *SeedIt);
	Trackcounter++;
	PropagateTrack(Track,SuperLayer-1);
      }
    }
  }
#ifdef L3_VERBOSE
  trk0 = nTracks();
  if(L3_Parameter::Trace()) cout << "L3_Silicon::TrackReconstructionPrime STOP, # of tracks " << trk0 << endl;
#endif 
}

void L3_Silicon::TrackReconstruction98() {
//Track reconstruction
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::TrackReconstruction98 START" << endl;
#endif 
  TrackList.reserve(3000);
  int Trackcounter =0;
  for (int layer=6; layer>=2; layer--) {
    vector<L3_Track> tmpTrackList;
    tmpTrackList.reserve(2000);
    for (int i=0; i<4; i++) {
      for (vector<L3_SpacePoint>::iterator SeedIt 
	     = (SpacePointList[layer][i]).begin();
	   SeedIt != (SpacePointList[layer][i]).end();
	   ++SeedIt) {
	L3_Track Track(Trackcounter, *SeedIt);
	Trackcounter++;
	//	cout << " Trackseed : " << Track << endl;
	int depth = 0;
	for (int j=0; j<4; j++) {
	  SearchNextSL98(Track, layer, j, depth, tmpTrackList);
	}
      }
    }
#ifdef L3_VERBOSE
    cout << " " << tmpTrackList.size() 
	 << " trackssegments starting in superlayer " << layer << endl;
#endif
    for (vector<L3_Track>::iterator TrackIt = tmpTrackList.begin();
	 TrackIt != tmpTrackList.end(); ++TrackIt) {
//      TrackIt->refit();
      TrackList.push_back(*TrackIt);
//      cout << Track;
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::TrackReconstruction98 STOP" << endl;
#endif 
}

int npointsOld(vector<L3_SpacePoint> &l,L3_Window &SwinMean,int show)
{
  int n = 0;
  for (vector<L3_SpacePoint>::iterator
	 SpIt = lower_bound( l.begin(),
			     l.end(),
			     SwinMean, RPWin_less() ) ;
       SpIt != lower_bound( l.begin(),
			    l.end(),
			    SwinMean, RPWin_greater() ) ; 
       ++SpIt)
    {
      if(show) cout << SpIt->x() << "\t" << SpIt->y() << endl;
      n++;
    }
  return n;
}

int npointsNew(vector<L3_SpacePoint> &l,L3_Window &SwinMean,int show)
{
  int from = 0;
  int to = -1;
  int n = l.size();
  if( n>0 )
    {
      int N = n+10;
      int low = 0;
      int high = n-1;      
      int guess=-1,oldguess;
      do {    
	oldguess = guess;
	guess = (low+high)/2;
	//cout << low << ' ' << guess << ' ' << high << endl;
	if( l[guess].x() < SwinMean.xmin() )	  
	  low = guess+1;	  
	else if( l[guess].x() > SwinMean.xmax() )	  
	  high = guess-1;	  
	else
	  {	
	    from = guess-1;
	    to = guess+1;
	    //cout << from << ' ' << guess << ' ' << to << endl;
	    for(;from>=0 && l[from].x()>=SwinMean.xmin();from--);
	    from++;
	    for(;to<n && l[to].x()<=SwinMean.xmax();to++);
	    to--;
	    //cout << from << ' ' << to << endl;
	    break;
	  }
	if(N==1) 
	  {
	    cout << "**** fishy start ****" << endl;
	    cout << low << ' ' << guess << ' ' << high << endl;
	    for(vector<L3_SpacePoint>::iterator i = l.begin();i!=l.end();i++)
	      {
		int t = i-l.begin();
		cout << t << "\t" << i->x() << "\t" << i->y() << endl;
	      }
	    cout << SwinMean << endl;	   
	    cout << "**** fishy ****" << endl;
	  }
      } while(--N>0 && guess!=oldguess);   
    }
  n = (from>to) ? 0 : (to-from+1);
  if(show)
    {
      for(int i=from;i<=to;i++)
	{
	  L3_SpacePoint *SpIt = &l[i];
	  cout << SpIt->x() << "\t" << SpIt->y() << endl;
	}
    }
  return n;
}

void L3_Silicon::findHits(int &from,int &to,vector<L3_SpacePoint> &l,L3_Window &SwinMean)
{
  int n = l.size();
  if( n>0 )
    {
      int N = n+10;
      int low = 0;
      int high = n-1;      
      int guess=-1,oldguess;
      do {    
	oldguess = guess;
	guess = (low+high)/2;
	//cout << low << ' ' << guess << ' ' << high << endl;
	if( l[guess].x() < SwinMean.xmin() )	  
	  low = guess+1;	  
	else if( l[guess].x() > SwinMean.xmax() )	  
	  high = guess-1;	  
	else
	  {	
	    from = guess-1;
	    to = guess+1;
	    //cout << from << ' ' << guess << ' ' << to << endl;
	    for(;from>=0 && l[from].x()>=SwinMean.xmin();from--);
	    from++;
	    for(;to<n && l[to].x()<=SwinMean.xmax();to++);
	    to--;
	    //cout << from << ' ' << to << endl;
	    break;
	  }
	if(N==1) 
	  {
	    cout << "**** fishy start ****" << endl;
	    cout << low << ' ' << guess << ' ' << high << endl;
	    for(vector<L3_SpacePoint>::iterator i = l.begin();i!=l.end();i++)
	      {
		int t = i-l.begin();
		cout << t << "\t" << i->x() << "\t" << i->y() << endl;
	      }
	    cout << SwinMean << endl;	   
	    cout << "**** fishy ****" << endl;
	  }
      } while(--N>0 && guess!=oldguess);   
    }
}

int L3_Silicon::findHits(vector<L3_SpacePoint> &l,L3_Window &SwinMean,vector<int> &IDs)
{
  float d1=0.0,d2=0.0;
  int nfound = 0;
  int n = l.size();
  if( n>0 )
    {
      int N = n+10;
      int low = 0;
      int high = n-1;      
      int guess=-1,oldguess;
      do {    
	oldguess = guess;
	guess = (low+high)/2;
	//cout << low << ' ' << guess << ' ' << high << endl;
	if( l[guess].x() < SwinMean.xmin() )	  
	  low = guess+1;	  
	else if( l[guess].x() > SwinMean.xmax() )	  
	  high = guess-1;	  
	else
	  {	
	    float diff = l[guess].x() - SwinMean.x();
	    if( diff >= 0.0 ){
	      //we are at larger x than the center, move towards lower x and check for
	      //points even closer to the searchwindow center
	      for(int i=guess-1;i>=0 && l[i].x()>=SwinMean.xmin();i--){
		float newdiff = fabs( l[i].x() - SwinMean.x() );
		if( newdiff <= diff ){
		  //a better choice
		  diff = newdiff;
		  guess = i;
		} else {
		  //even worse, stop searching 
		  break;
		}
	      }
	    } else {
	      //we are at lower x than the center, move towards higher x and check for
	      //points even closer to the searchwindow center
	      diff = fabs(diff);
	      for(int i=guess+1;i<n && l[i].x()<=SwinMean.xmax();i++){
		float newdiff = fabs( l[i].x() - SwinMean.x() );
		if( newdiff <= diff ){
		  //a better choice
		  diff = newdiff;
		  guess = i;
		} else {
		  //even worse, so stop searching
		  break;
		}
	      }	 
	    }   
	    if(SwinMean.inY( l[guess].y() )) IDs[nfound++] = guess;
	    int upper = guess+1;
	    int lower = guess-1;
	    while( nfound<(int)IDs.size() ){
	      int w = 0;
	      while(upper<n){
		if( l[upper].x() > SwinMean.xmax() ){
		  //out of range
		  upper = n;
		} else if( !SwinMean.inY( l[upper].y() ) ){
		  //spacepoint is NOT in y-range, try next
		  upper++;
		} else
		  break;
	      }
	      while(lower>=0){
		if( l[lower].x() < SwinMean.xmin() ){
		  //out of range
		  lower = -1;		  
		} else if( !SwinMean.inY( l[lower].y() ) ){
		  //spacepoint is NOT in y-range, try next
		  lower--;
		} else
		  break;
	      }
	      if( upper<n  ){ d1 = fabs( l[upper].x() - SwinMean.x() ); w|=1; }
              if( lower>=0 ){ d2 = fabs( l[lower].x() - SwinMean.x() ); w|=2; }	        
	      switch(w){	
	      case 0  : return nfound;
	      case 1  : IDs[nfound++]=upper++; break;
	      case 2  : IDs[nfound++]=lower--; break;	      
	      default : if( d1<d2 ) IDs[nfound++]=upper++; else IDs[nfound++]=lower--; break;
	      }
	    }
	    break;
	  }
	if(N==1) 
	  {
	    cout << "**** fishy start ****" << endl;
	    cout << low << ' ' << guess << ' ' << high << endl;
	    for(vector<L3_SpacePoint>::iterator i = l.begin();i!=l.end();i++)
	      {
		int t = i-l.begin();
		cout << t << "\t" << i->x() << "\t" << i->y() << endl;
	      }
	    cout << SwinMean << endl;	   
	    cout << "**** fishy ****" << endl;
	  }
      } while(--N>0 && guess!=oldguess);   
    }
  return nfound;
}

int L3_Silicon::SearchNextSL(L3_Track& Track, const int& layer, int depth){
  depth++;      
  int SLtoSearch = layer-depth;
  int Nr_of_Tracks_reconstructed = 0;
  L3_Window SearchWindow;

  for(int sec=0;sec<4;sec++){  //loop over all sectors

    if (depth > 1) {
      Track.fit( L3_Geometry::SLz(SLtoSearch, sec) );
      SearchWindow = Track.window();
    } else {
      SearchWindow = Track.window( L3_Geometry::SLz(SLtoSearch, sec) );
    }
#ifdef ULLI 
    //const L3_Window &w = L3_Geometry::getSector(SLtoSearch,sec);
    //if( w.z()>-10.0 && !w.overlap(SearchWindow)) continue; //skip this sector
    int from = 0;
    int to = -1;
    findHits(from,to,SpacePointList[SLtoSearch][sec],SearchWindow);
    for(int i=from;i<=to;i++){   //loop over all spacepoints found in window
      L3_SpacePoint *SpIt = &SpacePointList[SLtoSearch][sec][i];
#else
    for (vector<L3_SpacePoint>::iterator
	   SpIt = lower_bound( (SpacePointList[SLtoSearch][sec]).begin(),
			       (SpacePointList[SLtoSearch][sec]).end(),
			       SearchWindow, RPWin_less() ) ;
	 SpIt != lower_bound( (SpacePointList[SLtoSearch][sec]).begin(),
			      (SpacePointList[SLtoSearch][sec]).end(),
			      SearchWindow, RPWin_greater() ) ; 
	 ++SpIt) {
#endif
      if ( (SpIt->y() >= SearchWindow.ymin()) &&
	   (SpIt->y() <= SearchWindow.ymax()) ) {      
	L3_Track newTrack(Track);
	newTrack.add(*SpIt);
	if (SLtoSearch > 0) { // Noch weitere Superlage anzuschauen
	  int Nr = SearchNextSL(newTrack,layer,depth);	
	  if (Nr == 0) { // No Spacepoint in Searchwindow in Next Superlayer
	    if (depth > 1) {
	      newTrack.refit();
	      if( newTrack.chisqr() < newTrack.chisqrCut() ){
		TrackList.push_back(newTrack);
		Nr_of_Tracks_reconstructed++;
	      } 
	    }
	  }
	} else { // reached front of silicon or track touches 5 Superlayers
	  newTrack.refit();
	  if( newTrack.chisqr() < newTrack.chisqrCut() ){
	    TrackList.push_back(newTrack);
	    Nr_of_Tracks_reconstructed++;
	  }
	}
      }
    }
  }
  return Nr_of_Tracks_reconstructed;
}

int L3_Silicon::PropagateTrack(L3_Track& Track,int SLtoSearch){

  L3_Window SearchWindows[4];  
  int candidates;
  vector<int> max(6);

  while(SLtoSearch>=0) {

    candidates = 0; //count number of candidates that are derived from this track

    for(int sec=0;sec<4;sec++){  //loop over all sectors

      //At present, the sectors (1 & 3) and (2 & 4) are at the same z-position, so
      //there is no need to calculate the searchwindow every time
      int winIndex = L3_Geometry::sectorAtSameZ(SLtoSearch,sec);
      if( winIndex>=sec ){
	//the searchwindow was NOT calculated yet, so do it
	SearchWindows[winIndex] = Track.window( L3_Geometry::SLz(SLtoSearch,sec) );	
      }
      L3_Window& SearchWindow = SearchWindows[winIndex]; //shorthand notation

      int anz = findHits(SpacePointList[SLtoSearch][sec],SearchWindow,max);
      for(int k=0;k<anz;k++){   //loop over all spacepoints found in window
	L3_SpacePoint& Sp = SpacePointList[SLtoSearch][sec][ max[k] ];
	L3_Track newTrack(Track);
	newTrack.add(Sp);
	if( SLtoSearch>0 ){  //there is another superlayer to look at
	  newTrack.setFlag(0); //set counter for SL without SP to 0
	  newTrack.fit(0.0); //fit track using new spacepoint
	  PropagateTrack(newTrack,SLtoSearch-1);
	} else if(newTrack.numPunkte() >= 3){	    
	  newTrack.refit();
	  TrackList.push_back(newTrack);		    
	}
	candidates++;
      }
    }
    if( candidates > 0 ) return 0; //the track was prolonged

    if(--SLtoSearch>=0){ 
      if( Track.getFlag() == 1 ) //no spacepoint found in last superlayer
	SLtoSearch=-1; //give up
      else
	Track.setFlag(1); //mark that no spacepoint was found in this SL, but try next
    }	
  }   
  if( Track.numPunkte() >= 3 ){
    Track.refit();
    TrackList.push_back(Track);	
  }
  return 0;
}


int L3_Silicon::PropagateTrackOld(L3_Track& Track,int SLtoSearch){

  L3_Window SearchWindows[4];  
  int candidates;

  while(SLtoSearch>=0) {

    candidates = 0; //count number of candidates that are derived from this track

    for(int sec=0;sec<4;sec++){  //loop over all sectors

      //At present, the sectors (1 & 3) and (2 & 4) are at the same z-position, so
      //there is no need to calculate the searchwindow every time
      int winIndex = L3_Geometry::sectorAtSameZ(SLtoSearch,sec);
      if( winIndex>=sec ){
	//the searchwindow was NOT calculated yet, so do it
	SearchWindows[winIndex] = Track.window( L3_Geometry::SLz(SLtoSearch,sec) );	
      }
      L3_Window& SearchWindow = SearchWindows[winIndex]; //shorthand notation

      int from = 0;
      int to = -1;
      findHits(from,to,SpacePointList[SLtoSearch][sec],SearchWindow);
      for(int i=from;i<=to;i++){   //loop over all spacepoints found in window

	L3_SpacePoint *SpIt = &SpacePointList[SLtoSearch][sec][i];
	if ( (SpIt->y() >= SearchWindow.ymin()) &&
	     (SpIt->y() <= SearchWindow.ymax()) ) {      	
	  L3_Track newTrack(Track);
	  newTrack.add(*SpIt);
	  if( SLtoSearch>0 ){  //there is another superlayer to look at
	    newTrack.setFlag(0); //set counter for SL without SP to 0
	    newTrack.fit(0.0); //fit track using new spacepoint
	    PropagateTrack(newTrack,SLtoSearch-1);
	  } else if(newTrack.numPunkte() >= 3){	    
	    newTrack.refit();
	    TrackList.push_back(newTrack);		    
	  }
	  candidates++;
	}
      }
    }
    if( candidates > 0 ) return 0; //the track was prolonged

    if(--SLtoSearch>=0){ 
      if( Track.getFlag() == 1 ) //no spacepoint found in last superlayer
	SLtoSearch=-1; //give up
      else
	Track.setFlag(1); //mark that no spacepoint was found in this SL, but try next
    }	
  }   
  if( Track.numPunkte() >= 3 ){
    Track.refit();
    TrackList.push_back(Track);	
  }
  return 0;
}

int L3_Silicon::SearchNextSL98(L3_Track& Track, const int& layer, const int& sec,
			       int depth, vector<L3_Track>& tmpTrackList) {
  depth++;
  int SLtoSearch = layer-depth;
#ifdef L3_DEBUG_SILICON
  cout << " SLtoSearch = " << SLtoSearch << endl;
#endif
  int Nr_of_Tracks_reconstructed = 0;
//  cout << " SLtoSearch (z)= " << L3_Geometry::SLz((SLtoSearch),sec) << endl;
  if ( L3_Geometry::SLz(SLtoSearch, sec) == 0. ) {
    return 0;
  }
  L3_Window SwinMean = Track.window( L3_Geometry::SLz(SLtoSearch, sec) );
  vector<L3_SpacePoint>::iterator SPListbegin;
  vector<L3_SpacePoint>::iterator SPListend;
  SPListbegin = (SpacePointList[SLtoSearch][sec]).begin();
  SPListend = (SpacePointList[SLtoSearch][sec]).end();
#ifdef L3_ONLINE_MODE
  if ( SLtoSearch != 2) {
#endif
#ifdef L3_OFFLINE_MODE
  if ( SLtoSearch != 0) {
#endif
    SPListbegin = (SpacePoint2XList[SLtoSearch][sec]).begin();
    SPListend = (SpacePoint2XList[SLtoSearch][sec]).end();
  }
//   if ( SLtoSearch == 2 &&  (SpacePointList[SLtoSearch][sec]).size() == 0 ) {
//     if (  (SpacePoint2List[SLtoSearch][sec]).size() == 0 ) {
//       SPListbegin = (SpacePoint2XList[SLtoSearch][sec]).begin();
//       SPListend = (SpacePoint2XList[SLtoSearch][sec]).end();
//     } else {
//       SPListbegin = (SpacePoint2List[SLtoSearch][sec]).begin();
//       SPListend = (SpacePoint2List[SLtoSearch][sec]).end();
//     }
//   } else {
//     SPListbegin = (SpacePointList[SLtoSearch][sec]).begin();
//     SPListend = (SpacePointList[SLtoSearch][sec]).end();
//   }
//   int forks=0;
//   if ( SLtoSearch == 1 || SLtoSearch == 2 ) {
//     SPListbegin = (SpacePointList[SLtoSearch][sec]).begin();
//     SPListend = (SpacePointList[SLtoSearch][sec]).end();
//     distance( lower_bound( SPListbegin, SPListend, SwinMean, RPWin_less() ),
// 	      lower_bound( SPListbegin, SPListend, SwinMean, RPWin_greater() ),
// 	      forks);
//     if (forks == 0) {
//       SPListbegin = (SpacePoint2List[SLtoSearch][sec]).begin();
//       SPListend = (SpacePoint2List[SLtoSearch][sec]).end();
//       distance( lower_bound( SPListbegin, SPListend, SwinMean, RPWin_less() ),
// 		lower_bound( SPListbegin, SPListend, SwinMean, RPWin_greater() ),
// 		forks);
//       if (forks == 0) {
// 	SPListbegin = (SpacePoint2XList[SLtoSearch][sec]).begin();
// 	SPListend = (SpacePoint2XList[SLtoSearch][sec]).end();
//       }
//     }
//   } else {
//     SPListbegin = (SpacePointList[SLtoSearch][sec]).begin();
//     SPListend = (SpacePointList[SLtoSearch][sec]).end();
//   }
//   cout << "Window: xmin = " << SwinMean.xmin(); 
//   cout << "  xmax = " << SwinMean.xmax(); 
//   cout << "    ymin = " << SwinMean.ymin(); 
//   cout << "  ymax = " << SwinMean.ymax() << endl; 
//   for (vector<L3_SpacePoint>::iterator SpIt = SPListbegin;
//        SpIt != SPListend; ++SpIt) {
//     cout << " Spacpoint: " << *SpIt;
//   }
//   int forks = 0;
//   distance( lower_bound( SPListbegin, SPListend, SwinMean, RPWin_less() ),
// 	    lower_bound( SPListbegin, SPListend, SwinMean, RPWin_greater() ),
// 	    forks);
//   cout << " Number of Spacepoints in Window: " << forks << endl;
  int forks = 0;
  for (vector<L3_SpacePoint>::iterator SpIt 
	 = lower_bound( SPListbegin, SPListend, SwinMean, RPWin_less() );
       SpIt != lower_bound( SPListbegin, SPListend,
			    SwinMean, RPWin_greater() ); ++SpIt) {
    L3_Window Swin;
    if ( SpIt->numHits() < 4 ) {
      Swin = Track.window(SpIt->z());
    } else {
      Swin = SwinMean;
    }
    if ( (SpIt->y() >= Swin.ymin()) &&
	 (SpIt->y() <= Swin.ymax()) ) {
      L3_Track newTrack(Track);
      newTrack.add(*SpIt);
      forks++;
#ifdef L3_DEBUG_SILICON
      cout << "newTrack:" << newTrack << endl;
#endif
      if (SLtoSearch > 0) { // Noch weitere Superlage anzuschauen
	int Nr = 0;
	for (int nsec=0; nsec<4; nsec++) {
	  Nr += SearchNextSL98(newTrack, layer, nsec, depth, tmpTrackList);
	}
	if (Nr == 0) { // No Spacepoint in Searchwindow in Next Superlayer
	  if (depth > 1) {
	    newTrack.refit();
	    tmpTrackList.push_back(newTrack);
	    Nr_of_Tracks_reconstructed++;
	  }
	}
      } else { // reached front of silicon
	newTrack.refit();
	tmpTrackList.push_back(newTrack);
	Nr_of_Tracks_reconstructed++;
      }
    }
  }
  if (forks == 0) {
#ifdef L3_ONLINE_MODE
    if ( SLtoSearch != 2) {
#endif
#ifdef L3_OFFLINE_MODE
    if ( SLtoSearch != 0) {
#endif
      SPListbegin = (SpacePoint2List[SLtoSearch][sec]).begin();
      SPListend = (SpacePoint2List[SLtoSearch][sec]).end();
      for (vector<L3_SpacePoint>::iterator SpIt 
	     = lower_bound( SPListbegin, SPListend, SwinMean, RPWin_less() );
	   SpIt != lower_bound( SPListbegin, SPListend,
				SwinMean, RPWin_greater() ); ++SpIt) {
	L3_Window Swin = Track.window(SpIt->z());
	if ( (SpIt->y() >= Swin.ymin()) &&
	     (SpIt->y() <= Swin.ymax()) ) {
	  L3_Track newTrack(Track);
	  newTrack.add(*SpIt);
	  forks++;
	  if (SLtoSearch > 0) { // Noch weitere Superlage anzuschauen
	    int Nr = 0;
	    for (int nsec=0; nsec<4; nsec++) {
	      Nr += SearchNextSL98(newTrack, layer, nsec, depth, tmpTrackList);
	    }
	    if (Nr == 0) { // No Spacepoint in Searchwindow in next Superlayer
	      if (depth > 1) {
		newTrack.refit();
		tmpTrackList.push_back(newTrack);
		Nr_of_Tracks_reconstructed++;
	      }
	    }
	  } else { //reached front of silicon
	    newTrack.refit();
	    tmpTrackList.push_back(newTrack);
	    Nr_of_Tracks_reconstructed++;
	  }
	}
      }
    }
  }
  if (forks == 0) {
#ifdef L3_ONLINE_MODE
    if ( SLtoSearch != 2) {
#endif
#ifdef L3_OFFLINE_MODE
    if ( SLtoSearch != 0) {
#endif
      SPListbegin = (SpacePoint2XList[SLtoSearch][sec]).begin();
      SPListend = (SpacePoint2XList[SLtoSearch][sec]).end();
      for (vector<L3_SpacePoint>::iterator SpIt 
	     = lower_bound( SPListbegin, SPListend, SwinMean, RPWin_less() );
	   SpIt != lower_bound( SPListbegin, SPListend,
				SwinMean, RPWin_greater() ); ++SpIt) {
	L3_Window Swin = Track.window(SpIt->z());
	if ( (SpIt->y() >= Swin.ymin()) &&
	     (SpIt->y() <= Swin.ymax()) ) {
	  L3_Track newTrack(Track);
	  newTrack.add(*SpIt);
	  forks++;
	  if (SLtoSearch > 0) { // Noch weitere Superlage anzuschauen
	    int Nr = 0;
	    for (int nsec=0; nsec<4; nsec++) {
	      Nr += SearchNextSL98(newTrack, layer, nsec, depth, tmpTrackList);
	    }
	    if (Nr == 0) { // No Spacepoint in Searchwindow in Next Superlayer
	      if (depth > 1) {
		newTrack.refit();
		tmpTrackList.push_back(newTrack);
		Nr_of_Tracks_reconstructed++;
	      }
	    }
	  } else { // reached front of silicon
	    newTrack.refit();
	    tmpTrackList.push_back(newTrack);
	    Nr_of_Tracks_reconstructed++;
	  }
	}
      }
    }
  }
#ifdef L3_DEBUG_SILICON
  cout << " Nr of Tracks reconstr: " <<  Nr_of_Tracks_reconstructed << endl;
#endif
  return Nr_of_Tracks_reconstructed;
}

// TRACKinfo: Utility structure for L3_Silicon::findValidTracks
TRACKinfo::TRACKinfo() {}
TRACKinfo::TRACKinfo(double p, int n): the_probability( p ), the_numHits( n ) {}

// - Find a set of L3_Track's, which do not exclude each other too much -
// Returns number of reasonable tracks
// Arguments:
//   maxCommonHits: maximum number of shared hits of two tracks 5
//   maxProbDiff:   maximal probability loss allowed for additional hits 0.2
//   minProb:       minimal accepted probability 0.0001
void L3_Silicon::findValidTracks() {
  int maxCommonHits = L3_Parameter::maxCommonHits();
  float maxProbDiff = L3_Parameter::maxProbDiff();
  float minProb = L3_Parameter::minProb();
  vector<L3_Track> validTracks;
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::findValidTracks START" << endl;
#endif 
  map<L3_Track,TRACKinfo> info;
  for (vector<L3_Track>::iterator itrack = TrackList.begin();
       itrack != TrackList.end(); ++itrack ) {    
    float p = itrack->prob();
    if (p > minProb) {
      int maxSoFar = -1;
      L3_Track soFar;
      //find track with highest number of common hits for this candidate
      for ( vector<L3_Track>::iterator ifound = validTracks.begin();
	    ifound != validTracks.end(); ++ifound ) {
	if (fabs(itrack->mx() - ifound->mx()) < 0.02 ) {
	  if (fabs(itrack->my() - ifound->my()) < 0.02 ) {
	    int commonHits = itrack->getCommonHits(*ifound);
	    if(commonHits > maxSoFar) {
	      maxSoFar = commonHits;
	      soFar = *ifound;
	    }
	  }
	}
      }
      if (maxSoFar <= maxCommonHits) { //accept as a new track candidate
	info[*itrack] = TRACKinfo(p, itrack->numHits());
	validTracks.push_back(*itrack);
      } else if (info[*itrack].the_numHits > info[soFar].the_numHits) {  //replace 
	if (p + maxProbDiff > info[soFar].the_probability) {
	  *find(validTracks.begin(), validTracks.end(), soFar) = *itrack;
	}
      }
    }
  }
#if 0
  vector<TRACKinfo> info;
  for (vector<L3_Track>::iterator itrack = TrackList.begin();
       itrack != TrackList.end(); ++itrack ) {
    int nHits = itrack->numHits();
    float p = itrack->prob();
    if ( p > minProb ) {
      int maxSoFar = -1;
      int soFar =0;
      int count =0;
      for ( vector<L3_Track>::iterator iFound = validTracks.begin();
	    iFound != validTracks.end(); ++iFound ) {
	if (fabs(itrack->mx() - iFound->mx()) < 0.02) {
	  if (fabs(itrack->my() - iFound->my()) < 0.02) {
	    int commonHits = itrack->getCommonHits(*iFound);
	    if (commonHits > maxSoFar) {
	      maxSoFar = commonHits;
	      soFar = count;
	    }
	  }
	}
	count++;
      }
      if ( maxSoFar <= maxCommonHits ) {
	validTracks.push_back(*itrack);
	info.push_back(TRACKinfo(p, nHits));
      } else if ( (p + maxProbDiff) > info[soFar].the_probability
		  && (nHits > info[soFar].the_numHits) ) {
	validTracks[soFar] = *itrack;
      }
    }
  }
#endif
#ifdef L3_VERBOSE
  trk1 = validTracks.size();
  if(L3_Parameter::Trace()) cout << "L3_Silicon::findValidTracks STOP: #tracks = " << trk1 << endl;
#endif 
  TrackList = validTracks;
  //assign a number 1....N to each track
  for(vector<L3_Track>::iterator q = TrackList.begin();q!=TrackList.end();q++)
    {
      q->setnr( q - TrackList.begin() + 1 );
    }
}

void L3_Silicon::FillHits() {
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::FillHits" << endl;
#endif
  //When running in ARTE we get our hits from the HITB-Table
#ifdef L3_ARTE_FRAME
// GEPT workaround until silarte is fixed
  if( !ArteTable<MCIN>::booked() ){  //these are real data
    for (ArteTable<GEDE>::iterator PIt=ArteTable<GEDE>::begin(); 
	 PIt != ArteTable<GEDE>::end(); ++PIt ) {
      ArtePointer<GEDE>& Pl = *PIt;
      if (Pl->cmp == 1) {
	Pl->hit1 = ArtePointer<HITB>(0);
	Pl->hitl = ArtePointer<HITB>(-1);
      }
    }
  }
  if (ArteTable<HITB>::booked() && ArtePointer<HITB>::nrow()>0 ) {
    for (ArteTable<HITB>::iterator hitbIt=ArteTable<HITB>::begin(); 
	 hitbIt!=ArteTable<HITB>::end(); ++hitbIt ) {
      ArtePointer<HITB>& hitb = *hitbIt;
      ArtePointer<GEDE> Plane(hitb->gede);
      if (Plane->cmp == 1) {
	L3_Hitb hit(hitb);
	if( !ArteTable<MCIN>::booked() ){ //we are processing real data!
	  if (*(Plane->hit1) == ArtePointer<HITB>(0)) {
	    Plane->hit1 = ArtePointer<HITB>( hit.hit() );
	    Plane->hitl = ArtePointer<HITB>( hit.hit() );
	  }
	  if (*(Plane->hitl) < ArtePointer<HITB>(hit.hit()) ) {
	    Plane->hitl = ArtePointer<HITB>( hit.hit() );
	  }
	  if (*(Plane->hit1) > ArtePointer<HITB>(hit.hit()) ) {
	    Plane->hit1 = ArtePointer<HITB>( hit.hit() );
	  }
	}
	if (hit.z() < 180) {	  
	  HitbList[hit.gesl() -1][hit.sec() -1][hit.lay() -1].push_back(hit);
	}
      }
    }
  }
#endif  

#ifdef L3_SLP_FRAME
  int nhitb = 0;
  VDS_hitb* hitb = vdsslt_fillHITB(&nhitb);
  for (int i=0; i < nhitb; i++) {
    L3_Hitb hit(i, hitb[i]);
    if (hit.gesl() < 8 && hit.gesl()> -1) {
#ifdef L3_DEBUG_SILICON
      cout << "Hit: "   << hit.hit()  << "  "
	   << " gesl= " << hit.gesl() << " sec= "  << hit.sec()
	   << " lay= "  << hit.lay()  << " gede= " << hit.gede() << endl;
#endif 	  
      HitbList[hit.gesl() -1][hit.sec() -1][hit.lay() -1].push_back(hit);
    }
  }
#endif

//loop over all sectors, delete those with too many hits
  for (int i=0; i<7; i++) { //Locate Superlayer
    for (int j=0; j<4; j++) { //Locate Quadrant
      for (int k=0; k<4; k++) { //Locate Plane
	if ( (HitbList[i][j][k]).size() > 45 ) {
	  (HitbList[i][j][k]).clear();
	}
      }
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Silicon::FillHits: found " << nHitbs() << " hits." << endl;
#endif
}

void L3_Silicon::print() const {
#ifdef L3_VERBOSE
  //print(HList());
  //print(Points());
  //print(Tracks());
#endif
}

void L3_Silicon::printRecoSummary() const {
#ifdef L3_VERBOSE
  cout << "HITBs in SiVD            : " << nHitbs()  << endl; 
  cout << "Reconstructed Spacepoints: " << nPoints() << endl;
  cout << "Reconstructed Tracks     : " << nTracks() << endl;
#endif
}

void L3_Silicon::print(const vector<L3_Hitb>& HitbList) const {
#ifdef L3_VERBOSE
  cout << "Silicon Hit List" << endl;
  cout << "==================================================================="
       << endl;
  for (vector<L3_Hitb>::const_iterator HitIt = HitbList.begin();
       HitIt != HitbList.end(); ++HitIt) {
    cout << "Hit x: " << *HitIt << endl;
  }
#endif
}

void L3_Silicon::print(const vector<L3_SpacePoint>& SpacePointList) const {
#ifdef L3_VERBOSE
  cout << "Reconstructed Spacepoints" << endl;
  cout << "==================================================================="
       << endl;
  cout << "Nr.\t#Hits\t Layer1\t\t Layer2\t\t Layer3\t\t Layer4\t\t\t\tL3_SpacePoint" << endl;
  cout << "-------------------------------------------------------------------"
       << endl;
  for (vector<L3_SpacePoint>::const_iterator SpacePointListIt =
	 SpacePointList.begin(); SpacePointListIt != SpacePointList.end();
       ++SpacePointListIt) {
    int SpacePointNumber=0;
    distance(SpacePointList.begin(), SpacePointListIt, SpacePointNumber);
    if (SpacePointNumber <= 300) {
      cout << (SpacePointNumber+1);
      cout << *SpacePointListIt;
    }
  }
  cout << "*******************************************************************"
       << endl;
#endif
}

void L3_Silicon::print(const vector<L3_Track>& TrackList) const {
#ifdef L3_VERBOSE
  cout << "Reconstructed Tracks" << endl;
  cout << "==================================================================="
       << endl;
  cout << "Nr.\tTrack" << endl;
  cout << "-------------------------------------------------------------------"
       << endl;
  for (vector<L3_Track>::const_iterator SpurIt = TrackList.begin();
       SpurIt != TrackList.end(); ++SpurIt) {
    int TrackNumber=0;
    distance(TrackList.begin(), SpurIt, TrackNumber);
    if (TrackNumber <= 100) {
      cout << (TrackNumber+1);
      cout << *SpurIt;
      cout << endl;
    }
  }
  cout << "*******************************************************************"
       << endl;
#endif
}

vector<L3_Hitb> L3_Silicon::getUnusedHitbs() const {
  vector<L3_Hitb> result;
  for(int sl=0;sl<7;sl++){
    for(int sec=0;sec<4;sec++){
      const vector<L3_SpacePoint>& spc = SpacePointList[sl][sec];
      vector<L3_Hitb> hits; //all used hitbs in this sl,sec
      for(vector<L3_SpacePoint>::const_iterator i=spc.begin();i!=spc.end();i++){
	i->getCopyOfHitbs(hits);
      }
      for(int pl=0;pl<4;pl++){
	const vector<L3_Hitb>& hitb = HitbList[sl][sec][pl];	  	
	for(vector<L3_Hitb>::const_iterator i=hitb.begin();i!=hitb.end();i++){
	  if( find(hits.begin(),hits.end(),*i)==hits.end() ){
	    //this hit is not part of any spacepoint in this sl,sec
	    result.push_back(*i);
	  }
	}	
      }      
    }
  }
  return result;
}

#if 0
#include "L3_Histogram.hh"
void L3_Silicon::unused(){
  static L3_Tuple *t = NULL;
  if(t==NULL) t = L3_Histogram::BookNT("t",43);
  static L3_Tuple *tt = NULL;
  if(tt==NULL) tt = L3_Histogram::BookNT("tt",44);
  for(int sl=0;sl<7;sl++){
    int Used = 0;
    int notUsed = 0;
    for(int sec=0;sec<4;sec++){
      vector<L3_SpacePoint>& spc = SpacePointList[sl][sec];
      vector<L3_Hitb> hits;
      for(vector<L3_SpacePoint>::iterator i=spc.begin();i!=spc.end();i++){
	i->getCopyOfHitbs(hits);
      }
      for(int pl=0;pl<4;pl++){
	vector<L3_Hitb>& hitb = HitbList[sl][sec][pl];	  
	int flag = -1;
	for(vector<L3_Hitb>::iterator i=hitb.begin();i!=hitb.end();i++){
	  if( find(hits.begin(),hits.end(),*i)==hits.end() ){
	    notUsed++;
	    cout << "unused sl/sec/lay " <<sl+1<<'/'<<sec+1<<'/'<<pl+1<<' '<<i->hit()<<endl; 
	    flag = 0;
	  } else {
	    flag = 1;
	    Used++;
	  }
	  t->column("sl",sl+1);
	  t->column("sec",sec+1);
	  t->column("lay",pl+1);
	  t->column("flag",flag);
	  const L3_Plane &plane = L3_Geometry::getPlane(sl,sec,pl);
	  if( plane.l3nsl()!=-1 ){	    
	    if( fabs(plane.cos())>0.5 ){
	      float f[3]={plane.symn(),0.5*(plane.symn()+plane.symx()),plane.symx()};
	      for(int k=0;k<3;k++){
		float x = (i->x() + f[k]*plane.sin())/plane.cos();
		tt->column("sl",sl+1);
		tt->column("sec",sec+1);
		tt->column("lay",pl+1);
		tt->column("x",x);
		tt->column("y",f[k]);
		tt->column("f",1);	  
		tt->dumpData();
	      }  
	    } else {
	      float f[3]={plane.sxmn(),0.5*(plane.sxmn()+plane.sxmx()),plane.sxmx()};
	      for(int k=0;k<3;k++){
		float y = (f[k]*plane.cos()-i->x())/plane.sin();
		tt->column("sl",sl+1);
		tt->column("sec",sec+1);
		tt->column("lay",pl+1);
		tt->column("x",f[k]);
		tt->column("y",y);
		tt->column("f",1);
		tt->dumpData();
	      }  
	    }
	  }
	  t->dumpData();
	}	
      }      
    }
    cout << "SL " << sl+1 << ' ' << 100.0*Used/float(Used+notUsed) << endl;
  }
}
#endif
  
#ifdef L3_ARTE_FRAME
void L3_Silicon::fillTracksFromRseg(){
   int cnt=1; //track counter
   for(ArteTable<RSEG>::iterator i=ArteTable<RSEG>::begin();i!=ArteTable<RSEG>::end();i++){
     ArtePointer<RSEG> rseg(*i);
     //loop over all found track segments, find hits in svd, SL1-7
     vector<L3_Hitb> l;
     unsigned int SL = 0; //a bit mask indicating which superlayers have hits
     for ( ArteRelationIterator<HITB> hitbIt = rseg->hitb.begin(); hitbIt != rseg->hitb.end(); ++hitbIt ){
       ArtePointer<HITB> hit(*hitbIt);
       ArtePointer<GEDE> pgede(hit->gede);
       if( pgede->cmp==1 && hit->z<180.0 ){	 
	 L3_Hitb help(hit);
	 if( help.gesl()>=1 && help.gesl()<=7 ){
	   l.push_back( help );
	   SL |= 1 << help.gesl();
	 } else
	   cout << "hitb with gesl == " << help.gesl() << endl;
       }
     }
     int nSL = 0;
     { //count superlayers hit
       for(int i=1;i<8;i++) if(SL & (1<<i)) nSL++;       
     }
     if(L3_Parameter::Info()){
       cout << "track segment with " << l.size() << " SVD hits in " << nSL << " superlayers, "; 
       for(int i=1;i<8;i++) if(SL & (1<<i)) cout << '1'; else cout << '0';
       cout << endl;
     }
     if( nSL >= 3 ){
       //in our design, a track contains pointers to spacepoints, so we have to generate the
       //spacepoints first before we can create a track
       vector<L3_Hitb> hitList[7][4];
       for(vector<L3_Hitb>::iterator i=l.begin();i!=l.end();i++){
	 int sl = i->gesl();
	 int sec = i->sec();
	 if( sl>=1 && sl<=7 && sec>=1 && sec<=4 ){
	   hitList[sl-1][sec-1].push_back(*i);
	 } else {
	   cout << "L3_Silicon: error index error sl,sec " << sl << ',' << sec << endl;
	 }
       }
       vector<L3_SpacePoint*> points; //this will help us to create a track
       for(int sl=0;sl<7;sl++) for(int sec=0;sec<4;sec++) if( hitList[sl][sec].size() > 0 ){	  
	 L3_SpacePoint point(hitList[sl][sec]);
	 if( SpacePointList[sl][sec].size() == 0 ){
	   //this is a tricky thing: since we store pointers to elements in
	   //vector<L3_SpacePoint*> points we may not allow for any changes in the
	   //spacepoints
	   SpacePointList[sl][sec].reserve(500);
	 }
	 SpacePointList[sl][sec].push_back(point);
	 points.push_back( &SpacePointList[sl][sec].back() );
       }
       L3_Track track(cnt++,rseg->xf,rseg->yf,rseg->txf,rseg->tyf,rseg->zf,points,-1);
       TrackList.push_back( track );       
     }
   }
#if 0
   for(vector<L3_Track>::iterator i=TrackList.begin();i!=TrackList.end();i++){
     cout << "Track " << i->nr() << " ***********************" << endl;
     vector<L3_SpacePoint> vv = i->getPunktList();
     cout << vv.size() << " spacepoints" << endl;
     for(vector<L3_SpacePoint>::iterator r=vv.begin();r!=vv.end();r++){
       cout << *r << endl;
     }
   }
#endif
}
#endif

void L3_Silicon::getSpacepointInfo(int* a,int* b,int* c) const {
  if(a) *a = spc0;
  if(b) *b = spc1;
  if(c) *c = spc2;
}

void L3_Silicon::getTrackInfo(int* a,int* b,int* c) const {
  if(a) *a = trk0;
  if(b) *b = trk1;
  if(c) *c = trk2;
}






