// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: HERA-B track covariance matrix class
//              Attention: propagation makes transformation
//              sigma(1/p) -> sigma(p) (a Vt-requirement).
//
// changes:
// 21 Aug 2000 (TG) creation
//
// *****************************************************************************

#include <iostream>
#include <assert.h>
#include "vt++/CMatrix.hh"
#include "arte/arte.hh"

namespace MATRIX {
  //============================================================================
  // CMatrix Constructors
  //============================================================================
  CMatrix::CMatrix(const CMatrix& rhs):VtSymMatrix(rhs) {}

  CMatrix::CMatrix(const ArtePointer<RTRA>& rtra):VtSymMatrix(5) {
    set_x      (rtra->cvf[0]);
    set_x_y    (rtra->cvf[1]);
    set_y      (rtra->cvf[2]);
    set_x_tx   (rtra->cvf[3]);
    set_y_tx   (rtra->cvf[4]);
    set_tx     (rtra->cvf[5]);
    set_x_ty   (rtra->cvf[6]);
    set_y_ty   (rtra->cvf[7]);
    set_tx_ty  (rtra->cvf[8]);
    set_ty     (rtra->cvf[9]);
    set_x_pinv (rtra->cvf[10]);
    set_y_pinv (rtra->cvf[11]);
    set_tx_pinv(rtra->cvf[12]);
    set_ty_pinv(rtra->cvf[13]);
    set_pinv   (rtra->cvf[14]);

    // pinv might be 0 if no momentum is available for track
    if(x() <= 0. || y() <= 0. || tx() <= 0. || ty() <= 0. || pinv() < 0.) {
      cout << "CMatrix warning: corrupt covariance Matrix in RTRA " << rtra
	   << ": neg. or zero diagonal element!!!" << endl;
    }
  }

  CMatrix::CMatrix(const ArtePointer<RSEG>& rseg):VtSymMatrix(5) {
    set_x      (rseg->cvf[0]);
    set_x_y    (rseg->cvf[1]);
    set_y      (rseg->cvf[2]);
    set_x_tx   (rseg->cvf[3]);
    set_y_tx   (rseg->cvf[4]);
    set_tx     (rseg->cvf[5]);
    set_x_ty   (rseg->cvf[6]);
    set_y_ty   (rseg->cvf[7]);
    set_tx_ty  (rseg->cvf[8]);
    set_ty     (rseg->cvf[9]);
    set_x_pinv (rseg->cvf[10]);
    set_y_pinv (rseg->cvf[11]);
    set_tx_pinv(rseg->cvf[12]);
    set_ty_pinv(rseg->cvf[13]);
    set_pinv   (rseg->cvf[14]);

    // pinv might be 0 if no momentum is available for track
    if(x() <= 0. || y() <= 0. || tx() <= 0. || ty() <= 0. || pinv() < 0.) {
      cout << "CMatrix warning: corrupt covariance Matrix in RSEG " << rseg
	   << ": neg. or zero diagonal element!!!" << endl;
    }

  }

  CMatrix::CMatrix(const ArtePointer<MTRA>& mtra):VtSymMatrix(5) {
    set_x      (1e-9);
    set_x_y    (0.);
    set_y      (1e-9);
    set_x_tx   (0.);
    set_y_tx   (0.);
    set_tx     (1e-9);
    set_x_ty   (0.);
    set_y_ty   (0.);
    set_tx_ty  (0.);
    set_ty     (1e-9);
    set_x_pinv (0.);
    set_y_pinv (0.);
    set_tx_pinv(0.);
    set_ty_pinv(0.);
    set_pinv   (1e-9);
  }

  //============================================================================
  // CMatrix::propagate: propagate Matrix by dz
  //============================================================================
  void CMatrix::propagate(const double dz, const double p) {
    set_x      ( x_prop(dz) );
    set_x_y    ( x_y_prop(dz) );
    set_x_tx   ( x_tx_prop(dz) );
    set_x_ty   ( x_ty_prop(dz) );
    set_x_pinv ( x_pinv_prop(dz,p) );
    set_y      ( y_prop(dz) );
    set_y_tx   ( y_tx_prop(dz) );
    set_y_ty   ( y_ty_prop(dz) );
    set_y_pinv ( y_pinv_prop(dz,p) );
    set_tx_pinv( tx_pinv_prop(p) );
    set_ty_pinv( ty_pinv_prop(p) );
    set_pinv   ( pinv_prop(p) );
    // tx     is invariant
    // tx_ty  is invariant
    // ty     is invariant
    return;
  }

} // namespace MATRIX
