// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: an (incomplete) collection of distance functions
//
// changes:
// 21 Aug 2000 (TG) creation
//
// *****************************************************************************
#include <iostream>
#include "arte/arte.hh"
#include "vt++/VtUtil.hh"
#include "vt++/VtVertex.hh"
#include "vt++/VtTrack.hh"

namespace VERTEX {

  
  //============================================================================
  // distance: spatial distance between RVER and MVER
  //============================================================================
  double distance(const ArtePointer<RVER>& rver ,
		  const ArtePointer<MVER>& mver) {
    return sqrt(sqr(rver->xv - mver->x) + sqr(rver->yv - mver->y) +
		sqr(rver->zv - mver->z));
  }

  //============================================================================
  // distance: Chi2 distance between tracks
  //============================================================================
  double distance(Track& t1, Track& t2) {
    Vertex tmp(t1,t2);
    tmp.findVertexVt();
    return tmp.chi2();
  }

  //============================================================================
  // distance: spatial distance between Track and Vertex
  //============================================================================
  double distance(const Track& t, const Vertex& v) {
    double dx = v.vx() - t.x();
    double dy = v.vy() - t.y();
    double dz = v.vz() - t.z();
    double tx = t.tx();
    double ty = t.ty();
    double nom = sqr(dx*ty - dy*tx) + sqr(dy - dz*ty) + sqr(dz*tx - dx);
    double denom = sqr(tx) + sqr(ty) + 1.;
    return sqrt(nom/denom);
  }

  //============================================================================
  // distance: Chi2 distance between two vertices
  //============================================================================
  double distance(const Vertex& v1, const Vertex& v2) {
    return v1.distance(v2);
  }

  //============================================================================
  // distance: Chi2 distance between vertex and space point
  //============================================================================
  double distance(const Vertex& v, double x, double y, double z) {
    return v.distance(x,y,z);
  }


  //============================================================================
  // distance: spatial distance between Track and Wire
  //============================================================================
  double distance(const Track& t, const ArtePointer<GTAR>& gtar) {
    double mz1 = 1.;
    double rx2 = gtar->x;
    double ry2 = gtar->y;
    double rz2 = gtar->z;

    double drx = t.x() - rx2;
    double dry = t.y() - ry2;
    double drz = t.z() - rz2;
    double mx1 = t.tx();
    double my1 = t.ty();
    double mx2=0, my2=0, mz2=0;
    
    if(gtar->dx > gtar->dy) {
      mx2 = 1.;
      my2 = 0.;
      mz2 = 0.;
    } else {
      mx2 = 0.;
      my2 = 1.;
      mz2 = 0.;
    }

    double denom = sqr(mx1*my2-my1*mx2) + sqr(my1*mz2-mz1*my2) +
      sqr(mz1*mx2-mx1*mz2);

    double detx = drx*my1*mz2 - drx*mz1*my2;
    double dety = dry*mz1*mx2 - dry*mx1*mz2;
    double detz = drz*mx1*my2 - drz*my1*mx2;
    
    return fabs(detx+dety+detz)/denom;
  }

  //============================================================================
  // distance: spatial distance between RVER and GTAR
  //============================================================================
  double distance(const ArtePointer<RVER>& rver ,
		  const ArtePointer<GTAR>& gtar) {

    double x1 = gtar->x;
    double y1 = gtar->y;
    double z1 = gtar->z;

    double a, b;

    if(gtar->dx > gtar->dy) {
      a = 1.;
      b = 0.;
    } else {
      a = 0.;
      b = 1.;
    }
    
    double x3 = rver->xv;
    double y3 = rver->yv;
    double z3 = rver->zv;

    double nom = sqr((x3-x1)*b - (y3-y1)*a) + sqr((z3-z1)*b) + sqr((z3-z1)*a);
    double denom = sqr(a)+sqr(b);

    return sqrt(nom/denom);
  }

} // end of namespace VERTEX
