// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Vertex class
//
// changes:
// 21 Aug 2000 (TG) creation
//
// *****************************************************************************
#include <iostream>
#include <iomanip>
#include "arte/arte.hh"
#include "vt++/VtVertex.hh"
#include "vt++/VtTrack.hh"
#include "vt++/VtKalman.hh"
#include "vt++/VtRelation.hh"
#include "vt++/VtNegMatrix.hh"

using namespace MATRIX;

namespace VERTEX {

  static const VtSymMatrix CINV0(3,1.e-5);
  static const VtVector    kal_xv0(3);

  //==============================================================================
  // Constructors
  //==============================================================================
  Vertex::Vertex() :
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist (0.),
    v_bk13 (0.),
    v_chi2 (0.),
    kal_xv (VtVector(3)),
    kal_xvs(VtVector(3)),
    v_CS   (VtSymMatrix(3,1.e5))
    {}

  Vertex::Vertex(const ArtePointer<RVER>& v):
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist (0.),
    v_bk13 (0.),
    v_chi2 (0.),
    kal_xv (VtVector(3)),
    kal_xvs(VtVector(v->xv,v->yv,v->zv)),
    v_CS(3) {
    v_CS(0,0)             = v->cv[0];
    v_CS(0,1) = v_CS(1,0) = v->cv[1];
    v_CS(1,1)             = v->cv[2];
    v_CS(0,2) = v_CS(2,0) = v->cv[3];
    v_CS(1,2) = v_CS(2,1) = v->cv[4];
    v_CS(2,2)             = v->cv[5];

    // get the tracks
    for(ArteRelationIterator<RREL> it=v->rrel.begin(); it!=v->rrel.end(); ++it) {
      ArtePointer<RREL> rrel = *it;
      if(rrel->flag == 0) continue;
      Track* t = new Track(rrel->rtra);
      createRelation(*t, *this);
      tracks.push_back(t);
    }
  }
  
  Vertex::Vertex(const ArtePointer<RTRA>& r1,
		 const ArtePointer<RTRA>& r2):
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist (0.),
    v_bk13 (0.),
    v_chi2 (0.),
    kal_xv (VtVector(3)),
    kal_xvs(VtVector(3)),
    v_CS   (VtSymMatrix(3,1.e5)) {

    Track* t1 = new Track(r1);
    createRelation(*t1, *this);
    tracks.push_back(t1);

    Track* t2 = new Track(r2);
    createRelation(*t2, *this);
    tracks.push_back(t2);
  }
  
  Vertex::Vertex(const ArtePointer<RSEG>& r1,
		 const ArtePointer<RSEG>& r2):
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist (0.),
    v_bk13 (0.),
    v_chi2 (0.),
    kal_xv (VtVector(3)),
    kal_xvs(VtVector(3)),
    v_CS   (VtSymMatrix(3,1.e5)) {
    
    Track* t1 = new Track(r1);
    createRelation(*t1, *this);
    tracks.push_back(t1);

    Track* t2 = new Track(r2);
    createRelation(*t2, *this);
    tracks.push_back(t2);
  }

  Vertex::Vertex(const ArtePointer<MTRA>& m1,
		 const ArtePointer<MTRA>& m2):
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist (0.),
    v_bk13 (0.),
    v_chi2 (0.),
    kal_xv (VtVector(3)),
    kal_xvs(VtVector(3)),
    v_CS   (VtSymMatrix(3,1.e5)) {
    
    Track* t1 = new Track(m1);
    createRelation(*t1, *this);
    tracks.push_back(t1);

    Track* t2 = new Track(m2);
    createRelation(*t2, *this);
    tracks.push_back(t2);
  }
  
  Vertex::Vertex(Track& t1, Track& t2):
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist(0.),
    v_bk13(0.),
    v_chi2(0.),
    kal_xv(VtVector(3)),
    kal_xvs(VtVector(3)),
    v_CS(VtSymMatrix(3,1.e5)) {

    createRelation(t1, *this);
    createRelation(t2, *this);
  }

  Vertex::Vertex(const vector<Track*>& v):
    Track(),
    v_use_momentum(true),
    v_use_kalman(false),
    v_valid(false),
    v_mother(false),
    v_angle(0.),
    v_dist(0.),
    v_bk13(0.),
    v_chi2(0.),
    kal_xv(VtVector(3)),
    kal_xvs(VtVector(3)),
    v_CS(VtSymMatrix(3,1.e5)) {
    
    for(vector<Track*>::const_iterator it = v.begin(); it != v.end(); ++it) {
      createRelation(**it, *this);
    }
  }

  Vertex::Vertex(const Vertex& rhs):
    Track(rhs),
    v_use_momentum(rhs.v_use_momentum),
    v_use_kalman(false),
    v_valid(rhs.v_valid),
    v_mother(rhs.v_mother),
    v_angle(0.),
    v_dist(0.),
    v_bk13(0.),
    v_chi2(0.),
    kal_xv(rhs.kal_xv),
    kal_xvs(rhs.kal_xvs),
    v_CS(rhs.v_CS) {
    // do not copy relations
  }

  Vertex::~Vertex() {
    // delete tracks owned by vertex
    for(Track_cit it = tracks.begin(); it != tracks.end(); ++it) {
      delete *it;
    }
  }
  
  //==============================================================================
  // push_back
  //==============================================================================
  void Vertex::push_back(const ArtePointer<RTRA>& rtra) {
    push_back(*new Track(rtra));
  }

  void Vertex::push_back(const ArtePointer<RSEG>& rseg) {
    push_back(*new Track(rseg));
  }

  void Vertex::push_back(const ArtePointer<MTRA>& mtra) {
    push_back(*new Track(mtra));
  }

  void Vertex::push_back(Track& track) {
    createRelation(track, *this);
    v_valid = false;
  }

  //==============================================================================
  // remove: remove a relation
  //==============================================================================
  void Vertex::remove(Relation* r){ RelationList::remove(r); v_valid=false; }

  //==============================================================================
  // calc_mother: calculate mother particle properties
  //==============================================================================
  bool Vertex::calc_mother() {
    // add up track momenta
    t_p[5] = 0;
    double pxsum = 0., pysum = 0., pzsum = 0.;
    for(iterator it = begin(); it != end(); ++it) {
      const Kalman& t = it->kalman;
      double ez = 1./sqrt(1.+sqr(t.tx())+sqr(t.ty()));
      double ex = t.tx()*ez;
      double ey = t.ty()*ez;
      t_p[5] += t.p();
      pxsum  += t.p()*ex;
      pysum  += t.p()*ey;
      pzsum  += t.p()*ez;
    }

    if(fabs(pzsum)>1e-20) {
      t_p[3] = pxsum/pzsum; // tx
      t_p[4] = pysum/pzsum; // ty
    } else 
      return false;

    // x,y at z = 0
    t_p[0] = vx() - vz() * tx();
    t_p[1] = vy() - vz() * ty();
    t_p[2] = 0.;

    v_mother = true;
    //    t_p[5] = sqrt(sqr(pxsum) + sqr(pysum) + sqr(pzsum));

    // what is the difference betwenn sum of track.p() and
    // calculation sqrt(sqr(pxsum)+sqr(pysum)+sqr(pzsum))?????
    //      cout << " reco-p: " << sqrt(sqr(pxsum) + sqr(pysum) + sqr(pzsum))
    //    	 << " orig reco p: " << p() << endl;

    return true;
  }

  //==============================================================================
  // calc_mother: calculate mother particle covariance Matrix
  //==============================================================================
  bool Vertex::calc_mother_cov() {
    // no track parameters computed
    if(v_mother == false) return false;

    // build up big cov. matrix (in Vt: covn)
    // covn consists of vertex + track parameter cov. matrices:
    //  vtx  tr1  tr2  tr3  ...
    //
    //  CS | ES | ES | ES | ...   CC = cor(q_k,q_l) = F_k*C_n*F_l^T
    //  ---+----+----+----+          = -ES_k * F_l
    //  ES | DS | CC | CC | ...
    //  ---+----+----+----+ 
    //  ES | CC | DS | CC | ...
    //  ---+----+----+----+
    //  ES | CC | CC | DS | ...
    //  ...
    const unsigned int dim = 3*ntracks() + 3;
    VtSymMatrix covn(dim);

    covn.place_at(v_CS,0,0);

    unsigned int ipos = 3;
    for(iterator it = begin(); it != end(); ++it) {
      const Kalman& kalman = it->kalman;

      covn.place_at(kalman.ES(), 0, ipos);
      covn.place_at(kalman.DS(), ipos, ipos);

      // fill cor(q_k^n,q_l^n) = F_k*C_n*F_l^T = -F_k * E_l^n
      unsigned int jpos = 6;
      iterator tmp=it;
      for(iterator it2=++tmp; it2 != end(); ++it2) {
	const Kalman& kalman2 = it2->kalman;
	
	covn.place_at(-kalman.ES()*kalman2.F(), ipos,jpos);
	jpos+=3;
      }

      ipos+=3;
    }

    // fill jacobian
    VtMatrix    jac(dim,5);

    // mom unit vectors
    double epx = px()/p();
    double epy = py()/p();
    double epz = pz()/p();

    // vertex part:
    jac(0,0) = jac(1,1) = 1.;
    jac(2,0) = -tx();
    jac(2,1) = -ty();

    unsigned int i = 2;
    for(iterator it = begin(); it != end(); ++it) {
      const Kalman& t = it->kalman;
      
      // unit vectors along track:
      double ez = 1./sqrt(1. + sqr(t.tx()) + sqr(t.ty()));
      double ex = t.tx() * ez;
      double ey = t.ty() * ez;

      double fact2 = t.p() * ez;
      double fact3 = 1./(p() * epz);
      double fact  = fact2 * fact3;

      ++i;
      jac(i,2) = fact * (1. - sqr(ex) + tx() * ex * ez);
      jac(i,3) = fact * ex * (ty()*ez - ey);
      jac(i,0) = -vz() * jac(i,2);
      jac(i,1) = -vz() * jac(i,3);

      jac(i,4) = fact2 * ( epx*(1.-sqr(ex)) - epy*ex*ey - epz*ex*ez );
      
      ++i;
      jac(i,2) = fact*ey*(tx()*ez-ex);
      jac(i,3) = fact*(1. - sqr(ey) + ty()*ey*ez);
      jac(i,0) = -vz() * jac(i,2);
      jac(i,1) = -vz() * jac(i,3);

      jac(i,4) = fact2 * ( epy*(1.-sqr(ey)) - epx*ex*ey - epz*ey*ez );

      ++i;
      jac(i,2) = fact3 * (ex - tx()*ez);
      jac(i,3) = fact3 * (ey - ty()*ez);
      jac(i,0) = -vz() * jac(i,2);
      jac(i,1) = -vz() * jac(i,3);
      
      jac(i,4) = epx*ex + epy*ey * epz*ez;
    }

    // compute Track covariance matrix:
    t_V = covn.product(jac);
    // compute inverse
    t_G = t_V;
    t_G.VtDsinv();

    return true;
  }

  //==============================================================================
  // createRVER: save vertex in RVER table
  //==============================================================================
  const ArtePointer<RVER> Vertex::createRVER(void) const {
    RVER r;

    r.nt = ntracks();
    r.xv = vx();
    r.yv = vy();
    r.zv = vz();

    r.mass = 0.;
    r.merr = 0.;
    r.wire = 0;
    r.btra = 0;

    r.cv[0] = v_CS(0,0);
    r.cv[1] = v_CS(0,1);
    r.cv[2] = v_CS(1,1);
    r.cv[3] = v_CS(0,2);
    r.cv[4] = v_CS(1,2);
    r.cv[5] = v_CS(2,2);

    ArtePointer<RVER> rver = ArteTable<RVER>::push_back(r);

    double prchi2 = 0.;
    for(RelationList::const_iterator it = begin(); it != end(); ++it) {
      const Track& track = it->track;
      if(track.IsRTRA() == false) {
#ifdef VtDEBUG
	cout << "Vertex::createRVER error: track is not RTRA, cannot create RREL!" 
	     << endl;
#endif
	continue;
      }

      ArtePointer<RTRA> rtra = track.rtra();
      const Kalman& kalman = it->kalman;

      RREL rl;
      rl.chi  = kalman.chi2() - prchi2;
      rl.flag = 1;
      rl.topp = 0;

      prchi2 = kalman.chi2();

      ArtePointer<RREL> rrel = ArteTable<RREL>::push_back(rl);
      createRelation( rtra, rrel, &RTRA::rrel );
      createRelation( rver, rrel, &RVER::rrel );
    }

    return rver;
  }

  //==============================================================================
  // remove_last
  //==============================================================================
  const bool Vertex::remove_last() {
    return VtRemoveTrack(back());
  }

  //==============================================================================
  // distance: Chi2 distance to space point
  //==============================================================================
  double Vertex::distance(double xx, double yy, double zz) const {
    VtVector dx(xx-vx(),yy-vy(),zz-vz());
    return back()->kalman.CINV().product(dx);
  }


  //==============================================================================
  // distance: Chi2 distance to other vertex
  //==============================================================================
  double Vertex::distance(const Vertex& rhs) const {
    const VtSymMatrix& cov1 = back()->kalman.CINV();
    const VtSymMatrix& cov2 = rhs.back()->kalman.CINV();
    VtSymMatrix Cmat = cov1+cov2;

    if(Cmat.VtDsinv() == false) return 1000.;

    const VtVector x1  = cov1 * kal_xvs;
    const VtVector x2  = cov2 * rhs.kal_xvs;
    const VtVector xav = Cmat * (x1+x2);
    const VtVector dx1 = xav - kal_xvs;
    const VtVector dx2 = xav - rhs.kal_xvs;

    // get a chi2 value
    double sum1 = cov1.product(dx1);
    double sum2 = cov2.product(dx2);
    return (sum1 + sum2);
  }

  //==============================================================================
  // distance: Chi2 distance to track
  //==============================================================================
  double Vertex::distance(Track& t) const {
    Vertex tmp(*this);
    tmp.push_back(t);
    tmp.use_momentum(false);
    tmp.VtFilter();
    return tmp.chi2n();
  }

  //==============================================================================
  // chi2n
  //==============================================================================
  double Vertex::chi2n() const { 
    return back()->kalman.chi2();
  }

  //==============================================================================
  // print
  //==============================================================================
  void Vertex::print(std::ostream& os) const {
    os.setf(ios::right,ios::adjustfield);
    os.setf(ios::fixed,ios::floatfield);
    if(v_mother == true) {
      os << " Mother Track: ";
      Track::print(os);
      os << endl;
    }
    os << " Vertex:"
       << " vx: "       << setw(6) << setprecision(3) << vx()
       << " vy: "       << setw(6) << setprecision(3) << vy()
       << " vz: "       << setw(7) << setprecision(3) << vz()
       << " chi2: "     << setw(6) << setprecision(2) << chi2()
       << " angle: "    << setw(5) << setprecision(3) << angle() 
       << " dist: "     << setw(6) << setprecision(4) << dist()
       << " mass: "     << setw(6) << setprecision(4) << mass(v_use_kalman)
       << " no of tracks: " << ntracks() 
       << endl;

#ifdef VtDEBUG
    os << " Relations: ";

    RelationList::print(os);
#endif

    os << " Vertex cov. matrix: " << CS();

    return;
  }

  //==============================================================================
  // invariant mass functions
  //==============================================================================
  double Vertex::mass(const bool use) const {
    return mass(vector<double>(ntracks(),0), use);
  }

  double Vertex::mass(double m1) const {
    return mass(vector<double>(ntracks(),m1*m1));
  }

  double Vertex::mass(double m1, double m2) const {
    if(ntracks() != 2) return -1;
    vector<double> tmp(2);
    tmp[0] = m1*m1, tmp[1] = m2*m2;
    return mass(tmp);
  }

  double Vertex::mass(double m1, double m2, double m3) const {
    if(ntracks() != 3) return -1;
    vector<double> tmp(3);
    tmp[0] = m1*m1, tmp[1] = m2*m2, tmp[2] = m3*m3;
    return mass(tmp);
  }

  double Vertex::mass(double m1, double m2, double m3, double m4) const {
    if(ntracks() != 4) return -1;
    vector<double> tmp(4);
    tmp[0] = m1*m1, tmp[1] = m2*m2, tmp[2] = m3*m3, tmp[4] = m4*m4;
    return mass(tmp);
  }

  //==============================================================================
  // calc invariant mass
  //==============================================================================
  double Vertex::mass(const vector<double>& m02, const bool use) const {

    double esum  = 0.;
    double pxsum = 0.;
    double pysum = 0.;
    double pzsum = 0.;

    unsigned int i=0;
    for(RelationList::const_iterator it = begin(); it != end(); ++it) {
      const Track& t = it->track;
      double p;
      double tx;
      double ty;
      if(v_use_kalman == false && use == false) {
	tx = t.tx();
	ty = t.ty();
	p  = t.p();
      } else {
	const Kalman& k = it->kalman;
	tx = k.qvs()[0];
	ty = k.qvs()[1];
	p  = k.qvs()[2];
      }

      double pz = p / sqrt(tx*tx + ty*ty + 1.);
      double px = tx * pz;
      double py = ty * pz;
      pxsum += px;
      pysum += py;
      pzsum += pz;
      esum  += sqrt(p*p + m02[i++]);
    }

    double mass2 = esum*esum - pxsum*pxsum - pysum*pysum - pzsum*pzsum;
    if(mass2 < 0) return 0.;
    else return sqrt(mass2);
  }

  //==============================================================================
  // find 3D vertex of a set of tracks and return the rms distance from the vertex
  //==============================================================================
  const bool Vertex::findVertex3D(void) {

    if(size() < 2) return false;

    double c[3], v[3], tt[9];
    for(int i=0; i<9; i++) tt[i] = 0.;
    for(int i=0; i<3; i++) c[i]  = v[i] = 0.;

    double t[3], r[3], dt;
    for(RelationList::iterator pos = begin(); pos != end(); ++pos) {
      const Track& track = pos->track;

      t[0] = track.tx();  r[0] = track.x();
      t[1] = track.ty();  r[1] = track.y();
      t[2] = 1;           r[2] = track.z();
      
      double tsq = sqr(t[0]) + sqr(t[1]) + sqr(t[2]); 
      
      for(int i=0; i<3; i++) {
	for(int j=0; j<3; j++) {
	  dt = t[i]*t[j]/tsq; if(i==j) dt -= 1; 
	  c[i]      += dt*r[j];
	  tt[3*i+j] += dt;
	}
      }
    }
    
    int ifail = 0, ndim = 3;
    double work[3];
    dinv_(&ndim, tt, &ndim, work, &ifail);
    
    if(ifail) {
      cout << "findVertex: matrix inversion failed" << endl;
      return false;
    }
    
    // get the vertex position
    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) v[i] += tt[3*i+j]*c[j];
    }


    // fill vertex parameters
    kal_xvs[0]  = v[0];      // x-vertex
    kal_xvs[1]  = v[1];      // y-vertex
    kal_xvs[2]  = v[2];      // z-vertex

    // require that vertex-z is smaller than smallest hit-z of rsegs
    // to avoid vertexing the downstream end of rseg's
    for(RelationList::iterator it = begin(); it != end(); ++it) {
      if(vz() > it->track.z() ) return false;
    }

    v_valid = true;
    return true;
  }


  //=============================================================================
  // RMSdist: compute rms distance of tracks to vertex & rms opening angle
  //=============================================================================
  const double Vertex::rmsDistAngle(void) {

    // and the rms distance to the vertex
    v_dist = 0.;
    double tsq, rsq, rt, t[3], r[3];
    double sumtx = 0., sumty = 0.;
    double sumtx2 = 0., sumty2 = 0.;
    for(RelationList::iterator pos = begin(); pos != end(); ++pos) {
      const Track& track = pos->track;

      t[0] = track.tx();  r[0] = track.x() - kal_xvs[0];
      t[1] = track.ty();  r[1] = track.y() - kal_xvs[1];
      t[2] = 1;           r[2] = track.z() - kal_xvs[2];

      sumtx += t[0]; sumtx2 += sqr(t[0]);
      sumty += t[1]; sumty2 += sqr(t[1]);

      tsq = sqr(t[0]) + sqr(t[1]) + sqr(t[2]);
      rsq = sqr(r[0]) + sqr(r[1]) + sqr(r[2]);
      rt  = r[0]*t[0] + r[1]*t[1] + r[2]*t[2];

      v_dist += rsq - sqr(rt)/tsq;
    } 

    unsigned int v_nrseg = size();
    v_angle = sqrt(sumtx2*v_nrseg - sqr(sumtx) + sumty2*v_nrseg -
		   sqr(sumty))/static_cast<double>(v_nrseg); 
    
    v_dist /= static_cast<double>(v_nrseg-1);
    v_dist  = sqrt(v_dist);
    return v_dist;
  }

  //=============================================================================
  // find 2D vertex of a set of tracks and return the rms distance from the vertex
  //=============================================================================
  const bool Vertex::findVertex2D(void) {  

    unsigned int v_nrseg = size();
    if(v_nrseg < 2) return false;
    
    double sumtx=0, sumty=0, sumrx=0, sumry=0, sumrt=0, sumtt=0;

    double tx, ty, rx, ry;
    for(RelationList::iterator pos = begin(); pos != end(); ++pos) {
      const Track& track = pos->track;

      // slopes and position at z=0
      tx = track.tx();  rx = track.x() - tx*track.z();
      ty = track.ty();  ry = track.y() - ty*track.z();
      
      // update cumulants
      sumtx += tx;   sumrx += rx;
      sumty += ty;   sumry += ry;
      sumrt += rx*tx   + ry*ty;
      sumtt += sqr(tx) + sqr(ty);
    }

    // get vertex z-position

    double z = -(v_nrseg*sumrt - sumrx*sumtx - sumry*sumty) / 
                (v_nrseg*sumtt - sqr<double>(sumtx) - sqr<double>(sumty));
  
    // and the transverse coordinates
    double x = (sumrx + z*sumtx)/static_cast<double>(v_nrseg);
    double y = (sumry + z*sumty)/static_cast<double>(v_nrseg);

    // fill vertex parameters
    kal_xvs[0]  = x;         // x-vertex
    kal_xvs[1]  = y;         // y-vertex
    kal_xvs[2]  = z;         // z-vertex

    v_valid = true;
    return true;
  }

  //=============================================================================
  // VtFit: fit the vertex
  //=============================================================================
  const bool Vertex::VtFit() {
    if(VtFilter() == false) return false;
    else {
      VtSmoothX();
      VtSmoothQ();
    }
    return true;
  }

  //=============================================================================
  // findVertexVt: find Vertex with Vt package algorithm
  //=============================================================================
  const bool Vertex::findVertexVt() {  
    
    // 1st Step: find approximate Vertex from average of all given tracks.
    // Initialize Kalman-filter
    VtEstimateVertex();
    
#ifdef VtDEBUG
    cout << "-------------------------------" << endl;
    cout << "Estimated Vertex: (" << vx() << "," << vy() << "," << vz() << ")" << endl;
    cout << "Number of Tracks: " << size();
    copy(begin(), end(), ostream_iterator<Relation>(cout," "));
    cout << endl;
#endif

    double prChi2 = 0.;
    const Kalman& trackn = back()->kalman; // last element

    unsigned int iter = 0;
    do {

      if(VtFit() == false) return false;

      double chi2 = trackn.chi2();

#ifdef VtDEBUG
      cout << "----> chi2n: " << chi2 << " prChi2: " << prChi2 << " fabs(chi2-prChi2): " 
	   << fabs(chi2 - prChi2) << endl;
#endif

      if(fabs(chi2 - prChi2) > 1.) {
	prChi2 = chi2;
      } else
	break;
    } while(++iter < 20);

#ifdef VtDEBUG
    cout << "-------------------------------" << endl;
    cout << "No of iterations: " << iter << endl;
    cout << *this;
#endif

    if(iter<20) { //  && chi2()<10.
      v_valid = true;
      return true;
    } else {
      //      cout << "-----> VERTEX FIT FAILED!! <--------" << endl;
      return false;
    }
  }

  //=============================================================================
  // VtEstimateVertex (former VTAVER function)
  //=============================================================================
  const bool Vertex::VtEstimateVertex() {

    double tx_sum  = 0.;
    double x_sum   = 0.;
    double xw_sum  = 0.;
    double xtx_sum = 0.;
    double tx2_sum = 0.;
    double ty_sum  = 0.;
    double y_sum   = 0.;
    double yw_sum  = 0.;
    double yty_sum = 0.;
    double ty2_sum = 0.;

    // fill cumulants & initialize Kalman structure
    for(RelationList::iterator tit = begin(); tit != end(); ++tit) {
      const Track& track = tit->track;
      //      track->propagate(0);   // propagate tracks to z=0 (needed for Vt)
      // init some parameters needed for Kalman-Filter
      //      track->init_kalman(use_momentum, x(), y(), z());
      tit->kalman.set_use_momentum(v_use_momentum);

      double x        = track.x();
      double y        = track.y();
      double tx       = track.tx();
      double ty       = track.ty();
      double xweight  = 1./track.V().x();
      double yweight  = 1./track.V().y();
      double xweight2 = xweight*xweight;
      double yweight2 = yweight*yweight;

      tx_sum  += tx * xweight;
      x_sum   += x  * xweight;
      xw_sum  += xweight;
      xtx_sum += x  * tx * xweight2;
      tx2_sum += tx * tx * xweight2;

      ty_sum  += ty * yweight;
      y_sum   += y  * yweight;
      yw_sum  += yweight;
      yty_sum += y  * ty * yweight2;
      ty2_sum += ty * ty * yweight2;

    } // for track
    
    double det = -tx2_sum - ty2_sum + tx_sum*tx_sum/xw_sum + ty_sum*ty_sum/yw_sum;
    if(det == 0.) {
#ifdef VtDEBUG
      cout << "VtVertex::VtEstimateVertex error: determinant == 0.!" << endl;
#endif
      return false;
    }

    kal_xvs[2] = ( xtx_sum + yty_sum - tx_sum*x_sum/xw_sum - ty_sum*y_sum/yw_sum ) / det;
    kal_xvs[0] = ( x_sum + tx_sum * vz() ) / xw_sum;
    kal_xvs[1] = ( y_sum + ty_sum * vz() ) / yw_sum;

    v_valid = true;
    return true;
  }

  //=============================================================================
  // VtFilter (in Vt: vtfilt) 
  //=============================================================================
  const bool Vertex::VtFilter() {
    
    // initialize by vertex position & cov. matrix
    //    const VtVector*    prkal_xv = &kal_xv0;
    const VtVector*    prkal_xv = &kal_xvs;
    const VtSymMatrix* prCINV   = &CINV0;
    double             prChi2   = 0.;
    v_bk13 = vz();

    //    cout << "bk13: " << v_bk13 << endl;

    for(RelationList::iterator tit = begin(); tit != end(); ++tit) {
      Kalman& kalman = tit->kalman;

      //      cout << "----> FILTER TRACK " << track;
#ifdef VtDEBUG
      cout << endl;
#endif

      // compute contributions of track k
      if(kalman.filter(v_bk13, *prCINV, *prkal_xv) == false) {
	return false; // forget this vertex
      }

      kalman.filter_chi2(v_bk13, prChi2, *prCINV, *prkal_xv);
    
      //      cout << "Track-Kalman chi2: " << kalman.chi2() << endl;

      // store pointers to previous (k-1) values
      prkal_xv = &kalman.xv();
      prCINV   = &kalman.CINV();
      prChi2   = kalman.chi2();
    } // for tracks

    return true;
  }

  //=============================================================================
  // VtInverseFilter (in Vt: vtdis): get smoothed residuals for tracks
  //=============================================================================
  const bool Vertex::VtInverseFilter() const {
    
    // get last track (in Vt: x_n)
    const Kalman& trackn = back()->kalman;

    const VtSymMatrix* prCINV = &CINV0;

    for(RelationList::const_iterator tit = begin(); tit != end(); ++tit) {
      Kalman& kalman = tit->kalman;
      
      if(kalman.inverse_filter(v_bk13, trackn.CINV(), *prCINV,
			       trackn.xv()) == false) {
	return false;
      }
      prCINV   = &kalman.CINV();
    }

    return true;
  }

  //=============================================================================
  // VtInverseFilter (in Vt: vtkill): remove track from vertex
  //=============================================================================
  const bool Vertex::VtRemoveTrack(Relation* const trackk) {

    if(VtInverseFilter() == false) return false;

    // get last track (in Vt: x_n)
    const Kalman& trackn = back()->kalman;

    // compute chi2_n -> chi2_n - chi2_k,S (eq. 22)
    double dchi2 = trackn.chi2() - trackk->kalman.chi2s();
    
    // new smoothed vertex position
    kal_xvs = trackk->kalman.xnk();

    for(RelationList::iterator tit = begin(); tit != end(); ++tit) {
      // calculate qvs
      tit->kalman.calc_qvs(kal_xvs);
    }
    
    remove(trackk);

    v_chi2 = back()->kalman.set_chi2(dchi2);

    // refit the vertex
    return VtFit();
  }

  //=============================================================================
  // VtSmoothX (in Vt: vtsmo)  smooth fit parameters, update xvs
  //=============================================================================
  void Vertex::VtSmoothX() {
    // get last track (in Vt: x_n)
    const Kalman& kaln = back()->kalman;
    
#ifdef VtDEBUG
    cout << "--> VtSmooth called!" << endl;
#endif

    // update vertex position saved in xvs!!
    kal_xvs = kaln.xv(); 

#ifdef VtDEBUG
    cout << "kal_xvs: "    << kal_xvs  << endl;
    cout << "track->C(): " << kaln.C() << endl;
#endif

    // save smoothed covariance matrix of vertex
    v_CS = kaln.C();

    // save chi2
    v_chi2 = kaln.chi2();
    // divide by degrees of freedom
    v_chi2 /= 2*size()-3;

#ifdef VtDEBUG
    cout << "v_CS: " << v_CS << endl;
#endif

    return;
  }

  //=============================================================================
  // VtSmooth (in Vt: vtsmo)  smooth track parameters, update qvs
  //=============================================================================
  void Vertex::VtSmoothQ() {
   
#ifdef VtDEBUG
    cout << "---> VtSmoothQ called! <---" << endl;
#endif

    const Kalman& trackn = back()->kalman;
    for(RelationList::iterator tit = begin(); tit != end(); ++tit) {
      tit->kalman.smooth(v_bk13, kal_xvs, trackn.C());
    }
    
    return;
  }


  //=============================================================================
  // operator =
  //=============================================================================
  const Vertex& Vertex::operator=(const Vertex& rhs) {
    if( this == &rhs ) { return *this; }
    
    RelationList::operator=(rhs);

    v_use_momentum = rhs.v_use_momentum;
    v_use_kalman   = rhs.v_use_kalman;
    v_valid        = rhs.v_valid;
    v_angle        = rhs.v_angle;
    v_dist         = rhs.v_dist;
    v_bk13         = rhs.v_bk13;
    v_chi2         = rhs.v_chi2;

    kal_xv   = rhs.kal_xv;
    kal_xvs  = rhs.kal_xvs;
    v_CS     = rhs.v_CS;
    tracks   = rhs.tracks;

    return *this;
  }
  
  //=============================================================================
  // operator ==
  //=============================================================================
  const bool Vertex::operator==(const Vertex& rhs) const {
    return RelationList::operator==(rhs);
  }

  //=============================================================================
  // simple distance between vertices
  //=============================================================================
  const double Vertex::operator-(const Vertex& rhs) const {
    return sqrt(sqr(vx()-rhs.vx()) + sqr(vy()-rhs.vy()) + sqr(vz()-rhs.vz()));
  }


} // End of namespace

