#ifndef __CMATRIX_HH
#define __CMATRIX_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: HERA-B track covariance matrix class
//              Attention: propagation makes transformation
//              sigma(1/p) -> sigma(p) (a Vt-requirement).
//
// changes:
// 21 Aug 2000 (TG) creation
//
// *****************************************************************************

#include "VtSymMatrix.hh"

template <class T> class ArtePointer;
struct RTRA;
struct RSEG;
struct MTRA;

namespace MATRIX {
  class VtVector;

  template <class T>
  inline const double sgn(const T& x) { return (x==0)? 0. : (x<0)? -1. : 1.; }
 
  
  /** $5\times 5$ symmetric matrix class
      @memo HERA-B Track covariance matrix class
  */
  //============================================================================
  // Class CMatrix: Covariance Matrix class (symmetric,pos definite, dim=5) 
  //============================================================================
  class CMatrix : public VtSymMatrix {
  public:
    /** @name Constructors */
    ///
    CMatrix() : VtSymMatrix(5) {}
    ///
    CMatrix(const CMatrix& rhs);
    ///
    CMatrix(const VtSymMatrix& rhs) : VtSymMatrix(rhs) {}
    ///
    CMatrix(const VtSqMatrix& rhs)  : VtSymMatrix(rhs) {}
    ///
    CMatrix(const VtMatrix& rhs)    : VtSymMatrix(rhs) {}
    ///
    CMatrix(const ArtePointer<RTRA>& rtra);
    ///
    CMatrix(const ArtePointer<RSEG>& rseg);
    ///
    CMatrix(const ArtePointer<MTRA>& mtra);
    
    /** @name Access methods */
    /// rtra->cvf[0]
    double x()       const { return operator()(0,0); }
    /// rtra->cvf[1]
    double x_y()     const { return operator()(0,1); }
    /// rtra->cvf[3]
    double x_tx()    const { return operator()(0,2); }
    /// rtra->cvf[6]
    double x_ty()    const { return operator()(0,3); }
    /// rtra->cvf[10]
    double x_pinv()  const { return operator()(0,4); }
    /// rtra->cvf[2]
    double y()       const { return operator()(1,1); }
    /// rtra->cvf[4]
    double y_tx()    const { return operator()(1,2); }
    /// rtra->cvf[7]
    double y_ty()    const { return operator()(1,3); }
    /// rtra->cvf[11]
    double y_pinv()  const { return operator()(1,4); }
    /// rtra->cvf[5]
    double tx()      const { return operator()(2,2); }
    /// rtra->cvf[8]
    double tx_ty()   const { return operator()(2,3); }
    /// rtra->cvf[12]
    double tx_pinv() const { return operator()(2,4); }
    /// rtra->cvf[9]
    double ty()      const { return operator()(3,3); }
    /// rtra->cvf[13]
    double ty_pinv() const { return operator()(3,4); }
    /// rtra->cvf[14]
    double pinv()    const { return operator()(4,4); }

    /** @name Modification of matrix elements */
    ///
    void set_x      (const double x)      { operator()(0,0)                   = x;      }
    ///
    void set_x_y    (const double xy)     { operator()(1,0) = operator()(0,1) = xy;     }
    ///
    void set_x_tx   (const double xtx)    { operator()(0,2) = operator()(2,0) = xtx;    }
    ///
    void set_x_ty   (const double xty)    { operator()(0,3) = operator()(3,0) = xty;    }
    ///
    void set_x_pinv (const double xpinv)  { operator()(0,4) = operator()(4,0) = xpinv;  }
    ///
    void set_y      (const double y)      { operator()(1,1)                   = y;      }
    ///
    void set_y_tx   (const double ytx)    { operator()(1,2) = operator()(2,1) = ytx;    }
    ///
    void set_y_ty   (const double yty)    { operator()(1,3) = operator()(3,1) = yty;    }
    ///
    void set_y_pinv (const double ypinv)  { operator()(1,4) = operator()(4,1) = ypinv;  }
    ///
    void set_tx     (const double tx)     { operator()(2,2)                   = tx;     }
    ///
    void set_tx_ty  (const double txty)   { operator()(2,3) = operator()(3,2) = txty;   }
    ///
    void set_tx_pinv(const double txpinv) { operator()(2,4) = operator()(4,2) = txpinv; }
    ///
    void set_ty     (const double ty)     { operator()(3,3)                   = ty;     }
    ///
    void set_ty_pinv(const double typinv) { operator()(3,4) = operator()(4,3) = typinv; }
    ///
    void set_pinv   (const double pinv)   { operator()(4,4)                   = pinv;   }

    /** @name Error propagation */
    ///
    double x_prop      (double dz) const { return x() + 2. * dz * x_tx() + dz*dz * tx(); }
    ///
    double x_y_prop    (double dz) const { return x_y() + dz*(y_tx() + x_ty() + dz * tx_ty()); }
    ///
    double x_tx_prop   (double dz) const { return x_tx() + dz * tx(); }
    ///
    double x_ty_prop   (double dz) const { return x_ty() + dz * tx_ty(); }
    ///
    double x_pinv_prop (double dz, double p) const { return -sgn(p) * p*p * (x_pinv() + dz * tx_pinv()); }
    ///
    double y_prop      (double dz) const { return y() + 2. * dz * y_ty() + dz*dz * ty(); }
    ///
    double y_tx_prop   (double dz) const { return y_tx() + dz * tx_ty(); }
    ///
    double y_ty_prop   (double dz) const { return y_ty() + dz * ty(); }
    ///
    double y_pinv_prop (double dz, double p) const { return -sgn(p) * p*p * (y_pinv() + dz * ty_pinv() ); }
    ///
    double tx_pinv_prop(double p)  const { return -sgn(p) * tx_pinv() * p*p; }
    ///
    double ty_pinv_prop(double p)  const { return -sgn(p) * ty_pinv() * p*p; }
    ///
    double pinv_prop   (double p)  const { return pinv() * p*p*p*p;  }
    /// propagate all matrix elements
    void propagate(const double dz, const double p);

  }; // class CMatrix


} // namespace MATRIX
#endif
