#ifndef __VTTRACK_HH
#define __VTTRACK_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Track class
//
// changes:
// 21 Aug 2000 (TG) creation
//
// *****************************************************************************

#include <iosfwd>
#include "arte/arte.hh"
#include "vt++/VtUtil.hh"
#include "vt++/CMatrix.hh"
#include "vt++/VtVector.hh"
#include "vt++/VtRelationList.hh"

namespace VERTEX {
  class Kalman;

  /** Track representation class
   */
  //============================================================================
  // Class Track
  //============================================================================
  class Track: public RelationList {
  public:
    /**@name Constructors */
    ///
    Track();
    ///
    Track(const ArtePointer<RTRA>& rtra);
    ///
    Track(const ArtePointer<RSEG>& rseg);
    ///
    Track(const ArtePointer<MTRA>& mtra);
    ///
    Track(const Track* const track);
    ///
    Track(const Track& rhs);
    ///
    virtual ~Track();

    /**@name Access functions */
    /// x at z=0 (in Vt: p(3,..))
    double x()  const { return t_p[0]; }
    ///  y at z=0 (in Vt: p(4,..))
    double y()  const { return t_p[1]; }
    /// z = 0  in Vt
    double z()  const { return t_p[2]; }
    ///  slope    (in Vt: p(1,..))
    double tx() const { return t_p[3]; }
    ///  slope    (in Vt: p(2,..))
    double ty() const { return t_p[4]; }
    /// momentum (in Vt: p(5,..))
    double p()  const { return t_p[5]; }
    ///
    double px() const { return pz() * tx(); }
    ///
    double py() const { return pz() * ty(); }
    ///
    double pz() const { return p()/sqrt(1+tx()*tx()+ty()*ty()); }
    /// charge
    short int Q() const { return t_Q; }

    /// is Track constructed from RTRA object?
    bool   IsRTRA()              const { return t_isrtra; }
    /// $\sqrt{\sigma_x^2(z+dz)}$
    double xerr(double dz = 0)   const { return sqrt(t_V.x_prop(dz)); }
    /// $\sqrt{\sigma_y^2(z+dz)}$
    double yerr(double dz = 0)   const { return sqrt(t_V.y_prop(dz)); }
    /// $\sqrt{\sigma^2_{tx}}$
    double txerr()               const { return sqrt(t_V(2,2)); }
    /// $\sqrt{\sigma^2_{ty}}$
    double tyerr()               const { return sqrt(t_V(3,3)); }
    /// $\sqrt{\sigma^2_p}$
    double perr()                const { return sqrt(t_V.pinv_prop(p())); }

    /// return RSEG pointer
    ArtePointer<RSEG> rseg() const { return getPTR<RSEG>(t_arteptr); }
    /// return RTRA pointer
    ArtePointer<RTRA> rtra() const { return getPTR<RTRA>(t_arteptr); }
    /// return MTRA pointer
    ArtePointer<MTRA> mtra() const { return getPTR<MTRA>(t_arteptr); }

    /// covariance matrix
    const MATRIX::CMatrix&   V() const { return t_V; }
    /// inverse covariance matrix
    const MATRIX::CMatrix&   G() const { return t_G; }

    /**@name Expert functions */
    /// propagate track by $dz$
    void propagate(const double dz); 
    /// called by cout
    void print( std::ostream& os ) const;

    bool operator< (const Track& rhs) { return t_arteptr <  rhs.t_arteptr; }
    int  operator==(const Track& rhs) { return t_arteptr == rhs.t_arteptr; }

  private:
    unsigned int      t_arteptr;
    bool              t_isrtra;
    short int         t_Q;      // charge
  protected:
    MATRIX::VtVector  t_p;      // track parameter vector (x,y,z,tx,ty,p)
				// (in Vt: p(1,...,5) 
    MATRIX::CMatrix   t_V;      // 5x5 covariance matrix of Track
    MATRIX::CMatrix   t_G;      // inverse of t_cov
 }; // class Track


  //==============================================================================
  // operator <<
  //==============================================================================
  inline std::ostream& operator<< ( std::ostream& os, const Track& t ) {
    t.print(os);
    return os;
  }

} // end of namespace VERTEX
#endif
