#ifndef __VTVERTEX_HH
#define __VTVERTEX_HH
// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Vertex class
//
// changes:
// 21 Aug 2000 (TG) creation
//
// *****************************************************************************

#include <iosfwd>
#include <vector>
#include "CMatrix.hh"
#include "VtVector.hh"
#include "VtTrack.hh"

template <class T> class ArtePointer;
class RTRA;
class RSEG;
class MTRA;
class RVER;

namespace VERTEX {
  //
  // Some Declarations
  //
  class Track;
  class Vertex;

  typedef vector<Track*>                 Track_v;
  typedef vector<Track*>::iterator       Track_it;
  typedef vector<Track*>::const_iterator Track_cit;

  /** Vertex class */
  //============================================================================
  // Class Vertex
  //============================================================================
  class Vertex: public Track {
  public:
    
    /**@name Constructors */
    ///
    Vertex();
    ///
    Vertex(const Vertex& rhs);
    ///
    Vertex(const ArtePointer<RVER>& v);
    ///
    Vertex(const ArtePointer<RTRA>& r1, const ArtePointer<RTRA>& r2);
    ///
    Vertex(const ArtePointer<RSEG>& r1, const ArtePointer<RSEG>& r2);
    ///
    Vertex(const ArtePointer<MTRA>& m1, const ArtePointer<MTRA>& m2);
    ///
    Vertex(const vector<Track*>& v);
    ///
    Vertex(Track& t1, Track& t2);
    ///
    ~Vertex();

    /**@name Access functions */
    /// $x$ of vertex
    double vx()      const { return kal_xvs[0];  }
    /// $y$ of vertex
    double vy()      const { return kal_xvs[1];  }
    /// $z$ of vertex
    double vz()      const { return kal_xvs[2];  }
    /// available after call to rmsDistAngle()
    double angle()   const { return v_angle; }
    /// available after call to rmsDistAngle()
    double dist()    const { return v_dist;  } 
    /// no of tracks in vertex
    int    ntracks() const { return size(); }
    /// $\chi^2$ of vertex fit
    double chi2()    const { return v_chi2; }
    /// $\sqrt{\sigma_{vx}^2}$ vertex $x$-error
    double vxerr()   const { return sqrt(v_CS(0,0)); }
    /// $\sqrt{\sigma_{vy}^2}$ vertex $y$-error
    double vyerr()   const { return sqrt(v_CS(1,1)); }
    /// $\sqrt{\sigma_{vz}^2}$ vertex $z$-error
    double vzerr()   const { return sqrt(v_CS(2,2)); }
    /// $\sigma_{vx}^2$
    double vxerr2()  const { return v_CS(0,0); }
    /// $\sigma_{vy}^2$
    double vyerr2()  const { return v_CS(1,1); }
    /// $\sigma_{vz}^2$
    double vzerr2()  const { return v_CS(2,2); }

    /// $3\times3$ Vertex covariance matrix
    const MATRIX::VtSymMatrix&  CS() const { return v_CS; }
    /// is vertex valid?
    bool   valid()   const { return v_valid; }

    /** @name Mass calculation routines */
    /// rest-mass 0
    double mass(const bool use = false) const;
    /// single rest-mass calc.
    double mass(double m1) const;
    /// 2-track mass with particle rest-masses
    double mass(double m1, double m2) const;
    /// 3-track mass with rest-masses
    double mass(double m1, double m2, double m3) const;
    /// 4-track mass with rest-masses
    double mass(double m1, double m2, double m3, double m4) const;
    /// specify $n$ rest masses
    double mass(const vector<double>& m02, const bool use = false)  const;

    /**@name Vertexing routines */
    /// unweighted 2-D min. dist.
    const bool findVertex2D(void);
    /// unweighted 3-D min. dist.
    const bool findVertex3D(void);
    /// Vt Kalman-filter
    const bool findVertexVt    ();
    /// weighted 2-D min. dist.
    const bool VtEstimateVertex();


    /**@name Adding tracks */
    ///
    void   push_back(Track& track);
    ///
    void   push_back(const ArtePointer<RTRA>& rtra);
    ///
    void   push_back(const ArtePointer<RSEG>& rseg);
    ///
    void   push_back(const ArtePointer<MTRA>& mtra);


    /**@name Misc functions */
    /// write to RVER table
    const ArtePointer<RVER> createRVER(void) const;
    /// calc rms dist and rms angle
    const double rmsDistAngle();
    /// calc track params of mother particle
    bool calc_mother();
    /// calc cov. matrix of mother particle
    bool calc_mother_cov();

    /// remove last track from vertex, do inverse Kalman filter
    const bool remove_last();
    /// use track momentum for fit or not
    void  use_momentum(const bool use) { v_use_momentum = use; }
    ///
    bool  use_momentum() const { return v_use_momentum; }
    /// use refitted track parameters or not
    void  use_kalman(const bool use) { v_use_kalman = use; }
    ///
    bool  use_kalman() const { return v_use_kalman; }
    /// $\chi^2$ distance to space point
    double distance(double x, double y, double z) const;
    /// spatial distance to track
    double distance(Track& t) const;
    /// $\chi^2$ distance to vertex
    double distance(const Vertex& rhs) const;

    /**@name Expert routines */
    ///
    const bool VtFilter        ();
    ///
    const bool VtInverseFilter () const;
    ///
    void       VtSmoothX       ();
    ///
    void       VtSmoothQ       ();
    ///
    const bool VtRemoveTrack   (Relation* const track);
    ///
    const bool VtFit           ();
    ///
    const MATRIX::VtVector& xv()  const { return kal_xv; }
    /// vector of vertex-positions
    const MATRIX::VtVector& xvs() const { return kal_xvs; }
    ///
    void   print(std::ostream& os) const;
    /// $\chi^2$ contribution of last filter step
    double chi2n() const;
    /// unbook relation
    virtual void remove(Relation* r);
    
    //
    // Overloaded operators
    //
    const Vertex& operator= (const Vertex& rhs);
    const bool    operator==(const Vertex& rhs) const;
    const double  operator- (const Vertex& rhs) const;

  private:
    //    friend class VertexIterator;
    //    friend class ConstVertexIterator;
    bool   v_use_momentum;                // use momentum for vertex fit
    bool   v_use_kalman;                  // use refitted track parameters
    bool   v_valid;                       // is vertex information valid?
    bool   v_mother;                      // is the mother particle calculated?
    double v_angle;                       // rms of opening angle
    double v_dist;                        // rms of distance to vertex
    double v_bk13;                        // vtx-z at begin of iteration step
                                          // (in Vt: Bk13(iv), b13)
    double v_chi2;                        // chi2 of vertex-fit
    MATRIX::VtVector    kal_xv;           // x,y,z of vtx from kalman filter
    MATRIX::VtVector    kal_xvs;          // pos of Vertex after smooth
                                          // (xv-smoothed) (in Vt: xvs(1..3,...))
    MATRIX::VtSymMatrix v_CS;             // v_C smoothed (=v_C of last
                                          // filtered track) (in Vt: CS)
    Track_v    tracks;                    // tracks owned by vertex
  }; // class Vertex


  //==============================================================================
  // operator<<
  //==============================================================================
  inline std::ostream& operator<<(std::ostream& s, const Vertex& a) {
    a.print(s);
    return s;
  }

} // namespace VERTEX
#endif
